from datetime import datetime, timedelta

import xml.etree.ElementTree as ET
from flask_security import current_user

from opentakserver.extensions import db
from sqlalchemy import Integer, String, ForeignKey, Boolean, Float, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.forms.casevac_form import CasEvacForm
from opentakserver.functions import iso8601_string_from_datetime
from opentakserver.models.Icon import Icon


class CasEvac(db.Model):
    __tablename__ = 'casevac'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    sender_uid: Mapped[str] = mapped_column(String(255), ForeignKey("euds.uid", ondelete="CASCADE"))
    uid: Mapped[str] = mapped_column(String(255), unique=True)
    timestamp: Mapped[datetime] = mapped_column(DateTime)

    # The following are taken from CoT attributes generated from ATAK
    ambulatory: Mapped[int] = mapped_column(Integer, nullable=True)
    casevac: Mapped[bool] = mapped_column(Boolean, nullable=True)
    child: Mapped[int] = mapped_column(Integer, nullable=True)
    convenience: Mapped[int] = mapped_column(Integer, nullable=True)
    enemy: Mapped[str] = mapped_column(String(255), nullable=True)
    epw: Mapped[int] = mapped_column(Integer, nullable=True)
    equipment_detail: Mapped[str] = mapped_column(String(255), nullable=True)
    equipment_none: Mapped[bool] = mapped_column(Boolean, nullable=True)
    equipment_other: Mapped[bool] = mapped_column(Boolean, nullable=True)
    extraction_equipment: Mapped[bool] = mapped_column(Boolean, nullable=True)
    freq: Mapped[float] = mapped_column(Float, nullable=True)
    friendlies: Mapped[str] = mapped_column(String(255), nullable=True)
    hlz_marking: Mapped[int] = mapped_column(Integer, nullable=True)
    hlz_remarks: Mapped[str] = mapped_column(String(255), nullable=True)
    hoist: Mapped[bool] = mapped_column(Boolean, nullable=True)
    litter: Mapped[int] = mapped_column(Integer, nullable=True)
    marked_by: Mapped[str] = mapped_column(String(255), nullable=True)
    medline_remarks: Mapped[str] = mapped_column(String(255), nullable=True)
    nonus_civilian: Mapped[int] = mapped_column(Integer, nullable=True)
    nonus_military: Mapped[int] = mapped_column(Integer, nullable=True)
    obstacles: Mapped[str] = mapped_column(String(255), nullable=True)
    priority: Mapped[int] = mapped_column(Integer, nullable=True)
    routine: Mapped[int] = mapped_column(Integer, nullable=True)
    security: Mapped[int] = mapped_column(Integer, nullable=True)
    terrain_loose: Mapped[bool] = mapped_column(Boolean, nullable=True)
    terrain_other: Mapped[bool] = mapped_column(Boolean, nullable=True)
    terrain_other_detail: Mapped[bool] = mapped_column(String(255), nullable=True)
    terrain_detail: Mapped[str] = mapped_column(String(255), nullable=True)
    terrain_none: Mapped[bool] = mapped_column(Boolean, nullable=True)
    terrain_rough: Mapped[bool] = mapped_column(Boolean, nullable=True)
    terrain_slope: Mapped[bool] = mapped_column(Boolean, nullable=True)
    terrain_slope_dir: Mapped[str] = mapped_column(String(255), nullable=True)
    title: Mapped[str] = mapped_column(String(255))
    urgent: Mapped[int] = mapped_column(Integer, nullable=True)
    us_civilian: Mapped[int] = mapped_column(Integer, nullable=True)
    us_military: Mapped[int] = mapped_column(Integer, nullable=True)
    ventilator: Mapped[bool] = mapped_column(Boolean, nullable=True)
    winds_are_from: Mapped[str] = mapped_column(String(255), nullable=True)
    zone_prot_selection: Mapped[int] = mapped_column(Integer, nullable=True)
    urgent_surgical: Mapped[str] = mapped_column(String, nullable=True)
    point_id: Mapped[int] = mapped_column(Integer, ForeignKey("points.id", ondelete="CASCADE"), nullable=True)
    cot_id: Mapped[int] = mapped_column(Integer, ForeignKey("cot.id", ondelete="CASCADE"), nullable=True)
    point = relationship("Point", back_populates="casevac")
    cot = relationship("CoT", back_populates="casevac")
    zmist = relationship("ZMIST", back_populates="casevac", uselist=False, cascade="all, delete")
    eud = relationship("EUD", back_populates="casevacs")

    def from_wtforms(self, form: CasEvacForm):
        self.uid = form.uid.data
        self.timestamp = form.timestamp.data
        self.ambulatory = form.ambulatory.data
        self.casevac = form.casevac.data
        self.child = form.child.data
        self.convenience = form.convenience.data
        self.enemy = form.enemy.data
        self.epw = form.epw.data
        self.equipment_detail = form.equipment_detail.data
        self.equipment_none = form.equipment_none.data
        self.equipment_other = form.equipment_other.data
        self.extraction_equipment = form.extraction_equipment.data
        self.freq = form.freq.data
        self.friendlies = form.friendlies.data
        self.hlz_marking = form.hlz_marking.data
        self.hlz_remarks = form.hlz_remarks.data
        self.hoist = form.hoist.data
        self.litter = form.litter.data
        self.marked_by = form.marked_by.data
        self.medline_remarks = form.medline_remarks.data
        self.nonus_military = form.nonus_military.data
        self.nonus_civilian = form.nonus_civilian.data
        self.obstacles = form.obstacles.data
        self.priority = form.priority.data
        self.routine = form.routine.data
        self.security = form.security.data
        self.terrain_loose = form.terrain_loose.data
        self.terrain_other = form.terrain_other.data
        self.terrain_other_detail = form.terrain_other_detail.data
        self.terrain_detail = form.terrain_detail.data
        self.terrain_none = form.terrain_none.data
        self.terrain_rough = form.terrain_rough.data
        self.terrain_slope = form.terrain_slope.data
        self.terrain_slope_dir = form.terrain_slope_dir.data
        self.title = form.title.data
        self.urgent = form.urgent.data
        self.urgent_surgical = form.urgent_surgical.data
        self.us_military = form.us_military.data
        self.us_civilian = form.us_civilian.data
        self.ventilator = form.ventilator.data
        self.winds_are_from = form.winds_are_from.data
        self.zone_prot_selection = form.zone_prot_selection.data

    def serialize(self):
        return {
            'sender_uid': self.sender_uid,
            'uid': self.uid,
            'timestamp': self.timestamp,
            'ambulatory': self.ambulatory,
            'casevac': self.casevac,
            'child': self.child,
            'convenience': self.convenience,
            'enemy': self.enemy,
            'epw': self.epw,
            'equipment_detail': self.equipment_detail,
            'equipment_none': self.equipment_none,
            'equipment_other': self.equipment_other,
            'extraction_equipment': self.extraction_equipment,
            'freq': self.freq,
            'friendlies': self.friendlies,
            'hlz_marking': self.hlz_marking,
            'hlz_remarks': self.hlz_remarks,
            'hoist': self.hoist,
            'litter': self.litter,
            'marked_by': self.marked_by,
            'medline_remarks': self.medline_remarks,
            'nonus_civilian': self.nonus_civilian,
            'nonus_military': self.nonus_military,
            'obstacles': self.obstacles,
            'priority': self.priority,
            'routine': self.routine,
            'security': self.security,
            'terrain_loose': self.terrain_loose,
            'terrain_other': self.terrain_other,
            'terrain_other_detail': self.terrain_other_detail,
            'terrain_detail': self.terrain_detail,
            'terrain_none': self.terrain_none,
            'terrain_rough': self.terrain_rough,
            'terrain_slope': self.terrain_slope,
            'terrain_slope_dir': self.terrain_slope_dir,
            'title': self.title,
            'urgent': self.urgent,
            'urgent_surgical': self.urgent_surgical,
            'us_civilian': self.us_civilian,
            'us_military': self.us_military,
            'ventilator': self.ventilator,
            'winds_are_from': self.winds_are_from,
            'zone_prot_selection': self.zone_prot_selection,
            'eud': self.eud,
        }

    def to_json(self):
        icon = db.session.execute(db.session.query(Icon).filter(Icon.filename == 'red_crs.png')).first()[0]
        return {
            'sender_uid': self.sender_uid,
            'uid': self.uid,
            'timestamp': iso8601_string_from_datetime(self.timestamp),
            'ambulatory': self.ambulatory,
            'casevac': self.casevac,
            'child': self.child,
            'convenience': self.convenience,
            'enemy': self.enemy,
            'epw': self.epw,
            'equipment_detail': self.equipment_detail,
            'equipment_none': self.equipment_none,
            'equipment_other': self.equipment_other,
            'extraction_equipment': self.extraction_equipment,
            'freq': self.freq,
            'friendlies': self.friendlies,
            'hlz_marking': self.hlz_marking,
            'hlz_remarks': self.hlz_remarks,
            'hoist': self.hoist,
            'litter': self.litter,
            'marked_by': self.marked_by,
            'medline_remarks': self.medline_remarks,
            'nonus_civilian': self.nonus_civilian,
            'nonus_military': self.nonus_military,
            'obstacles': self.obstacles,
            'priority': self.priority,
            'routine': self.routine,
            'security': self.security,
            'terrain_loose': self.terrain_loose,
            'terrain_other': self.terrain_other,
            'terrain_other_detail': self.terrain_other_detail,
            'terrain_detail': self.terrain_detail,
            'terrain_none': self.terrain_none,
            'terrain_rough': self.terrain_rough,
            'terrain_slope': self.terrain_slope,
            'terrain_slope_dir': self.terrain_slope_dir,
            'title': self.title,
            'urgent': self.urgent,
            'urgent_surgical': self.urgent_surgical,
            'us_civilian': self.us_civilian,
            'us_military': self.us_military,
            'ventilator': self.ventilator,
            'winds_are_from': self.winds_are_from,
            'zone_prot_selection': self.zone_prot_selection,
            'zmist': self.zmist.serialize() if self.zmist else None,
            'eud': self.eud.to_json() if self.eud else None,
            'point': self.point.to_json() if self.point else None,
            'icon': icon.to_json(),
            'start': iso8601_string_from_datetime(self.cot.start) if self.cot else None,
            'stale': iso8601_string_from_datetime(self.cot.stale) if self.cot else None,
        }

    def to_cot(self):
        event = ET.Element("event")
        event.set("type", "b-r-f-h-c")
        event.set("version", "2.0")
        event.set("how", "h-g-i-g-o")
        event.set("uid", self.uid)
        event.set("time", iso8601_string_from_datetime(self.point.timestamp))
        event.set("start", iso8601_string_from_datetime(self.point.timestamp))
        event.set("stale", iso8601_string_from_datetime(self.point.timestamp + timedelta(days=365)))

        cot_point = ET.SubElement(event, "point")
        cot_point.set("ce", str(self.point.ce))
        cot_point.set("hae", str(self.point.hae))
        cot_point.set("le", str(self.point.le))
        cot_point.set("lat", str(self.point.latitude))
        cot_point.set("lon", str(self.point.longitude))

        detail = ET.SubElement(event, "detail")

        status = ET.SubElement(detail, "status")
        status.set("readiness", "false")

        contact = ET.SubElement(detail, "contact")
        contact.set("callsign", current_user.username)

        link = ET.SubElement(detail, "link")
        link.set("parent_callsign", current_user.username)
        link.set("production_time", iso8601_string_from_datetime(self.point.timestamp))
        link.set("relation", "p-p")
        link.set("type", "b-r-f-h-c")
        link.set("uid", self.uid)

        medevac = ET.SubElement(detail, "_medevac_")
        medevac.set("title", self.title)
        if self.casevac is not None:
            medevac.set("casevac", str(self.casevac).lower())
        if self.child is not None:
            medevac.set("child", str(self.child))
        if self.convenience is not None:
            medevac.set("convenience", str(self.convenience))
        if self.enemy is not None:
            medevac.set("enemy", str(self.enemy))
        if self.friendlies is not None:
            medevac.set("friendlies", str(self.friendlies))
        if self.nonus_military is not None:
            medevac.set("nonus_military", str(self.nonus_military))
        if self.nonus_civilian is not None:
            medevac.set("nonus_civilian", str(self.nonus_civilian))
        if self.us_civilian is not None:
            medevac.set("us_civilian", str(self.us_civilian))
        if self.us_military is not None:
            medevac.set("us_military", str(self.us_military))
        if self.epw is not None:
            medevac.set("epw", str(self.epw))
        if self.freq != None:
            medevac.set("freq", str(self.freq))
        if self.extraction_equipment is not None:
            medevac.set("extraction_equipment", str(self.extraction_equipment).lower())
        if self.equipment_detail:
            medevac.set("equipment_detail", self.equipment_detail)
        if self.equipment_other is not None:
            medevac.set("equipment_other", str(self.equipment_other).lower())
        if self.equipment_none is not None:
            medevac.set("equipment_none", str(self.equipment_none).lower())
        if self.hoist is not None:
            medevac.set("hoist", str(self.hoist).lower())
        if self.litter is not None:
            medevac.set("litter", str(self.litter))
        if self.marked_by:
            medevac.set("marked_by", self.marked_by)
        if self.hlz_marking is not None:
            medevac.set("hlz_marking", str(self.hlz_marking))
        if self.hlz_remarks:
            medevac.set("hlz_remarks", self.hlz_remarks)
        if self.medline_remarks:
            medevac.set("medline_remarks", self.medline_remarks)
        if self.security is not None:
            medevac.set("security", str(self.security))
        if self.terrain_detail:
            medevac.set("terrain_detail", self.terrain_detail)
        if self.terrain_other_detail is not None:
            medevac.set("terrain_other_detail", str(self.terrain_other_detail).lower())
        if self.terrain_loose is not None:
            medevac.set("terrain_loose", str(self.terrain_loose).lower())
        if self.terrain_none is not None:
            medevac.set("terrain_none", str(self.terrain_none).lower())
        if self.terrain_rough is not None:
            medevac.set("terrain_rough", str(self.terrain_rough).lower())
        if self.terrain_slope is not None:
            medevac.set("terrain_slope", str(self.terrain_slope).lower())
        if self.terrain_slope_dir:
            medevac.set("terrain_slope_dir", self.terrain_slope_dir)
        if self.obstacles:
            medevac.set("obstacles", self.obstacles)
        if self.priority is not None:
            medevac.set("priority", str(self.priority))
        if self.routine is not None:
            medevac.set("routine", str(self.routine))
        if self.urgent is not None:
            medevac.set("urgent", str(self.urgent))
        if self.urgent_surgical is not None:
            medevac.set("urgent_surgical", str(self.urgent_surgical))
        if self.ventilator is not None:
            medevac.set("ventilator", str(self.ventilator).lower())
        if self.winds_are_from:
            medevac.set("winds_are_from", self.winds_are_from)
        if self.zone_prot_selection is not None:
            medevac.set("zone_prot_selection", str(self.zone_prot_selection))

        if self.zmist:
            zmist_map = ET.SubElement(medevac, "zMistsMap")
            zmist_tag = ET.SubElement(zmist_map, "zMist")
            if self.zmist.i:
                zmist_tag.set("i", self.zmist.i)
            else:
                zmist_tag.set("i", "")
            if self.zmist.m:
                zmist_tag.set("m", self.zmist.m)
            else:
                zmist_tag.set("m", "")
            if self.zmist.s:
                zmist_tag.set("s", self.zmist.s)
            else:
                zmist_tag.set("s", "")
            if self.zmist.t:
                zmist_tag.set("t", self.zmist.t)
            else:
                zmist_tag.set("t", "")
            if self.zmist.title:
                zmist_tag.set("title", self.zmist.title)
            else:
                zmist_tag.set("title", "ZMIST1")
            if self.zmist.z:
                zmist_tag.set("z", self.zmist.z)
            else:
                zmist_tag.set("z", "")

        return event



"""Fixes for the eud_stats table

Revision ID: c9fa07339363
Revises: 85dbe0a7dcc1
Create Date: 2025-03-21 15:01:54.874938

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'c9fa07339363'
down_revision = '85dbe0a7dcc1'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table('eud_stats', schema=None) as batch_op:
        batch_op.alter_column('heap_free_size',
               existing_type=sa.INTEGER(),
               type_=sa.BigInteger(),
               existing_nullable=True)
        batch_op.alter_column('storage_total',
               existing_type=sa.INTEGER(),
               type_=sa.BigInteger(),
               existing_nullable=True)
        batch_op.alter_column('deviceDataRx',
               existing_type=sa.INTEGER(),
               type_=sa.BigInteger(),
               existing_nullable=True)
        batch_op.alter_column('heap_max_size',
               existing_type=sa.INTEGER(),
               type_=sa.BigInteger(),
               existing_nullable=True)
        batch_op.alter_column('storage_available',
               existing_type=sa.INTEGER(),
               type_=sa.BigInteger(),
               existing_nullable=True)
        batch_op.alter_column('deviceDataTx',
               existing_type=sa.INTEGER(),
               type_=sa.BigInteger(),
               existing_nullable=True)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table('eud_stats', schema=None) as batch_op:
        batch_op.alter_column('deviceDataTx',
               existing_type=sa.BigInteger(),
               type_=sa.INTEGER(),
               existing_nullable=True)
        batch_op.alter_column('storage_available',
               existing_type=sa.BigInteger(),
               type_=sa.INTEGER(),
               existing_nullable=True)
        batch_op.alter_column('heap_max_size',
               existing_type=sa.BigInteger(),
               type_=sa.INTEGER(),
               existing_nullable=True)
        batch_op.alter_column('deviceDataRx',
               existing_type=sa.BigInteger(),
               type_=sa.INTEGER(),
               existing_nullable=True)
        batch_op.alter_column('storage_total',
               existing_type=sa.BigInteger(),
               type_=sa.INTEGER(),
               existing_nullable=True)
        batch_op.alter_column('heap_free_size',
               existing_type=sa.BigInteger(),
               type_=sa.INTEGER(),
               existing_nullable=True)

    # ### end Alembic commands ###



"""Added mission_uids table

Revision ID: 807c3ca8e7d0
Revises: bde915ea136e
Create Date: 2024-10-11 16:53:44.954118

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision = '807c3ca8e7d0'
down_revision = 'bde915ea136e'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('mission_uids',
    sa.Column('uid', sa.String(length=255), nullable=False),
    sa.Column('mission_name', sa.String(length=255), nullable=True),
    sa.Column('timestamp', sa.DateTime(), nullable=True),
    sa.Column('creator_uid', sa.String(length=255), nullable=True),
    sa.Column('cot_type', sa.String(length=255), nullable=True),
    sa.Column('callsign', sa.String(length=255), nullable=True),
    sa.Column('iconset_path', sa.String(length=255), nullable=True),
    sa.Column('color', sa.String(length=255), nullable=True),
    sa.Column('latitude', sa.Float(), nullable=True),
    sa.Column('longitude', sa.Float(), nullable=True),
    sa.ForeignKeyConstraint(['mission_name'], ['missions.name'], ),
    sa.PrimaryKeyConstraint('uid')
    )
    op.drop_table('mission_logs')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('mission_logs',
    sa.Column('id', sa.INTEGER(), nullable=False),
    sa.Column('content', sa.String(length=255), nullable=False),
    sa.Column('creator_uid', sa.String(length=255), nullable=False),
    sa.Column('entry_uid', sa.String(length=255), nullable=False),
    sa.Column('server_time', sa.DATETIME(), nullable=False),
    sa.Column('dtg', sa.DATETIME(), nullable=False),
    sa.Column('created', sa.DATETIME(), nullable=False),
    sa.Column('keywords', sqlite.JSON(), nullable=False),
    sa.Column('mission_name', sa.String(length=255), nullable=True),
    sa.Column('content_hash', sa.String(length=255), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.drop_table('mission_uids')
    # ### end Alembic commands ###



import datetime
import uuid
from dataclasses import dataclass
from xml.etree.ElementTree import Element, SubElement

from opentakserver.functions import iso8601_string_from_datetime
from opentakserver.extensions import db
from sqlalchemy import Integer, String, Boolean, ForeignKey, DateTime, JSON
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.models.MissionChange import MissionChange


@dataclass
class MissionLogEntry(db.Model):
    __tablename__ = "mission_logs"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    content: Mapped[str] = mapped_column(String(255))
    creator_uid: Mapped[str] = mapped_column(String(255))
    entry_uid: Mapped[str] = mapped_column(String(255), default=str(uuid.uuid4()))
    mission_name: Mapped[str] = mapped_column(String(255), ForeignKey("missions.name"))
    server_time: Mapped[datetime] = mapped_column(DateTime, default=datetime.datetime.now(datetime.timezone.utc))
    dtg: Mapped[datetime] = mapped_column(DateTime, default=datetime.datetime.now(datetime.timezone.utc))
    created: Mapped[datetime] = mapped_column(DateTime, default=datetime.datetime.now(datetime.timezone.utc))
    # Leaving content_hash in order to not break existing DBs where this field is populated
    content_hash: Mapped[str] = mapped_column(String(255), nullable=True)
    content_hashes: Mapped[JSON] = mapped_column(JSON, default=[], nullable=True)
    keywords: Mapped[JSON] = mapped_column(JSON, default=[])
    mission = relationship("Mission", back_populates="mission_logs")

    def serialize(self):
        return {
            'content': self.content,
            'creator_uid': self.creator_uid,
            'entry_uid': self.entry_uid,
            'mission_names': self.mission_name,
            'server_time': self.server_time,
            'dtg': self.dtg,
            'created': self.created,
            'content_hashes': self.content_hashes,
            'keywords': self.keywords
        }

    def to_json(self):
        return {
            'id': self.entry_uid,
            'content': self.content,
            'creatorUid': self.creator_uid,
            'entryUid': self.entry_uid,
            'missionNames': [self.mission_name],
            'servertime': iso8601_string_from_datetime(self.server_time),
            'dtg': iso8601_string_from_datetime(self.dtg),
            'created': iso8601_string_from_datetime(self.created),
            'contentHashes': self.content_hashes if self.content_hash else [],
            'keywords': self.keywords if self.keywords is not None else []
        }

    def generate_cot(self) -> Element:
        event = Element("event", {"how": "h-g-i-g-o", "type": "t-x-m-c-l", "version": "2.0", "uid": str(uuid.uuid4()),
                                  "start": iso8601_string_from_datetime(self.dtg),
                                  "time": iso8601_string_from_datetime(self.server_time),
                                  "stale": iso8601_string_from_datetime(self.dtg + datetime.timedelta(minutes=2)),
                                  "access": "Undefined"})

        SubElement(event, "point", {"ce": "9999999", "le": "9999999", "hae": "0.0", "lat": "0.0", "lon": "0.0"})

        detail = SubElement(event, "detail")

        SubElement(detail, "mission", {"type": MissionChange.CHANGE, "tool": "public", "name": self.mission.name,
                                       "guid": self.mission.guid, "authorUid": self.creator_uid})

        return event



"""Added distro column to plugin table

Revision ID: b0c97a4ce0df
Revises: f310052349cb
Create Date: 2025-05-02 21:27:56.708927

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'b0c97a4ce0df'
down_revision = 'f310052349cb'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('plugins', schema=None) as batch_op:
        batch_op.add_column(sa.Column('distro', sa.String(length=255), nullable=False))
        batch_op.create_unique_constraint("distro", ['distro'])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('plugins', schema=None) as batch_op:
        batch_op.drop_constraint("distro", type_='unique')
        batch_op.drop_column('distro')

    # ### end Alembic commands ###



import tldextract
from flask_security import MailUtil
from opentakserver.extensions import logger


class EmailValidator(MailUtil):

    def __init__(self, app):
        super().__init__(app)
        self.app = app

    def validate(self, email: str) -> str:
        domain_whitelist = self.app.config.get("OTS_EMAIL_DOMAIN_WHITELIST")
        domain_blacklist = self.app.config.get("OTS_EMAIL_DOMAIN_BLACKLIST")
        tld_whitelist = self.app.config.get("OTS_EMAIL_TLD_WHITELIST")
        tld_blacklist = self.app.config.get("OTS_EMAIL_TLD_BLACKLIST")

        domain = email.split("@")[-1]
        parsed_domain = tldextract.extract(domain)
        logger.debug("Got domain: {}".format(domain))

        if domain_whitelist and domain not in domain_whitelist:
            logger.error("Domain {} is not whitelisted".format(domain))
            raise ValueError("Domain {} is not whitelisted".format(domain))

        if domain_blacklist and domain in domain_blacklist:
            logger.error("Domain {} is blacklisted".format(domain))
            raise ValueError("Domain {} is blacklisted".format(domain))

        if tld_whitelist and parsed_domain.suffix not in tld_whitelist:
            logger.error("TLD {} not whitelisted".format(parsed_domain.suffix))
            raise ValueError("TLD {} not whitelisted".format(parsed_domain.suffix))

        if tld_blacklist and parsed_domain.suffix in tld_blacklist:
            logger.error("TLD {} is blacklisted".format(parsed_domain.suffix))
            raise ValueError("TLD {} is blacklisted".format(parsed_domain.suffix))

        logger.info("Looks like {} is good".format(email))
        return super().validate(email)



import base64
import hashlib
import os.path
from datetime import datetime, timezone
from pathlib import Path

from bs4 import BeautifulSoup
from androguard.core.apk import APK

from werkzeug.utils import secure_filename

from sqlalchemy import Integer, String, Boolean, LargeBinary, DateTime
from sqlalchemy.orm import Mapped, mapped_column

from flask import current_app as app, request

from opentakserver.extensions import db
from opentakserver.forms.package_form import PackageForm
from opentakserver.functions import iso8601_string_from_datetime

from xml.etree.ElementTree import tostring


class Packages(db.Model):
    __tablename__ = "packages"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    platform: Mapped[str] = mapped_column(String(255))
    plugin_type: Mapped[str] = mapped_column(String(255))
    package_name: Mapped[str] = mapped_column(String(255))
    name: Mapped[str] = mapped_column(String(255))
    file_name: Mapped[str] = mapped_column(String(255))
    version: Mapped[str] = mapped_column(String(255))
    revision_code: Mapped[int] = mapped_column(Integer, nullable=True)
    description: Mapped[str] = mapped_column(String(255), nullable=True)
    apk_hash: Mapped[str] = mapped_column(String(255), nullable=True)
    os_requirement: Mapped[str] = mapped_column(Integer, nullable=True)
    tak_prereq: Mapped[str] = mapped_column(Integer, nullable=True)
    file_size: Mapped[int] = mapped_column(Integer)
    icon: Mapped[bytes] = mapped_column(LargeBinary, nullable=True)
    icon_filename: Mapped[str] = mapped_column(String(255), nullable=True)
    install_on_enrollment: Mapped[bool] = mapped_column(Boolean, default=False, nullable=True)
    install_on_connection: Mapped[bool] = mapped_column(Boolean, default=False, nullable=True)
    publish_time: Mapped[datetime] = mapped_column(DateTime)
    atak_version: Mapped[str] = mapped_column(String(255), nullable=True)

    def from_wtform(self, form: PackageForm):
        self.platform = form.platform.data
        self.plugin_type = form.plugin_type.data
        self.file_name = secure_filename(form.apk.data.filename)
        apk = APK(os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", self.file_name))
        self.package_name = apk.get_package()
        self.version = apk.get_androidversion_name()
        self.revision_code = apk.get_androidversion_code()
        self.os_requirement = apk.get_min_sdk_version()
        self.name = apk.get_app_name()
        self.description = form.description.data
        self.apk_hash = hashlib.sha256(form.apk.data.stream.read()).hexdigest()
        self.file_size = Path(os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", self.file_name)).stat().st_size
        self.icon = request.files['icon'].stream.read() if 'icon' in request.files else None
        self.icon_filename = secure_filename(request.files['icon'].filename) if 'icon' in request.files else None
        self.install_on_enrollment = form.install_on_enrollment.data
        self.install_on_connection = form.install_on_connection.data
        self.publish_time = datetime.now(timezone.utc)
        self.atak_version = form.atak_version.data

        manifest = BeautifulSoup(tostring(apk.get_android_manifest_xml()).decode('utf-8'))
        meta_data = manifest.find_all("meta-data", "lxml")
        for meta in meta_data:
            if 'ns0:value' not in meta.attrs or 'ns0:name' not in meta.attrs:
                continue

            if meta.attrs['ns0:name'] == "plugin-api":
                self.tak_prereq = meta.attrs['ns0:value']
                break

        if not self.icon:
            icon_filename, icon_extension = os.path.splitext(apk.get_app_icon())
            if icon_extension == '.png':
                self.icon = apk.get_file(apk.get_app_icon())
                self.icon_filename = f"{self.package_name}.png"

    def from_tak_gov(self, plugin: dict):
        self.apk_hash = plugin['apk_hash']
        self.file_size = plugin['apk_size_bytes']

        self.platform = plugin['platform']

    def serialize(self):
        return {
            'platform': self.platform,
            'plugin_type': self.plugin_type,
            'package_name': self.package_name,
            'name': self.name,
            'file_name': self.file_name,
            'version': self.version,
            'revision_code': self.revision_code,
            'description': self.description,
            'apk_hash': self.apk_hash,
            'os_requirement': self.os_requirement,
            'tak_prereq': self.tak_prereq,
            'file_size': self.file_size,
            'icon': self.icon,
            'icon_filename': self.icon_filename,
            'install_on_enrollment': self.install_on_enrollment,
            'install_on_connection': self.install_on_connection,
            'publish_time': self.publish_time,
            'atak_version': self.atak_version,
        }

    def to_json(self):
        return_value = self.serialize()
        return_value['publish_time'] = iso8601_string_from_datetime(self.publish_time)
        return_value['icon'] = f"data:image/png;base64,{base64.b64encode(self.icon).decode('utf-8')}" if self.icon else None
        return return_value



import csv
import datetime
import os
import zipfile

import bleach
import sqlalchemy.exc
from flask import current_app as app, request, Blueprint, jsonify, send_from_directory
from flask_security import roles_accepted, auth_required
from werkzeug.datastructures import ImmutableMultiDict

from opentakserver.blueprints.marti_api.marti_api import verify_client_cert
from opentakserver.forms.package_form import PackageForm, PackageUpdateForm
from opentakserver.models.Packages import Packages
from opentakserver.extensions import db, logger
from opentakserver.blueprints.ots_api.api import search, paginate
from opentakserver.blueprints.marti_api.certificate_enrollment_api import basic_auth

from werkzeug.utils import secure_filename

packages_blueprint = Blueprint('packages_api_blueprint', __name__)


@packages_blueprint.route('/api/packages/<package_name>')
@packages_blueprint.route('/api/packages/<atak_version>/<package_name>')
def download_package(package_name, atak_version=None):
    cert = verify_client_cert()
    if not cert:
        return '', 401
    return send_from_directory(os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages"), secure_filename(package_name))


@packages_blueprint.route('/api/packages')
@auth_required()
@roles_accepted("administrator")
def get_packages():
    query = db.session.query(Packages)
    query = search(query, Packages, 'platform')
    query = search(query, Packages, 'plugin_type')
    query = search(query, Packages, 'package_name')
    query = search(query, Packages, 'name')
    query = search(query, Packages, 'file_name')
    query = search(query, Packages, 'version')
    query = search(query, Packages, 'revision_code')
    query = search(query, Packages, 'apk_hash')
    query = search(query, Packages, 'tak_prereq')
    query = search(query, Packages, 'file_size')
    query = search(query, Packages, 'atak_version')

    return paginate(query)


@packages_blueprint.route('/api/packages/product.infz', methods=['HEAD'])
def head_product_infz():
    cert = verify_client_cert()
    if not cert:
        return '', 401

    if os.path.exists(os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", "product.infz")):
        return jsonify({'success': True})

    return jsonify({'success': False}), 404


@packages_blueprint.route('/api/packages/<atak_version>/product.infz', methods=['HEAD'])
def head_product_infz_with_version(atak_version: str):
    cert = verify_client_cert()
    if not cert:
        return '', 401

    if os.path.exists(os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", atak_version, "product.infz")):
        return jsonify({'success': True})

    return jsonify({'success': False}), 404


@packages_blueprint.route('/api/packages/product.infz', methods=['GET'])
def get_product_infz():
    cert = verify_client_cert()
    if not cert:
        return '', 401
    return send_from_directory(os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages"), "product.infz")


@packages_blueprint.route('/api/packages/<atak_version>/product.infz', methods=['GET'])
def get_product_infz_with_version(atak_version: str):
    cert = verify_client_cert()
    if not cert:
        return '', 401
    return send_from_directory(os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", atak_version), "product.infz")


def create_product_infz(atak_version: str | None):
    product_infz_file = os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", "product.infz")
    product_inf_file = os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", "product.inf")

    if atak_version:
        atak_version = bleach.clean(atak_version)
        product_infz_file = os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", atak_version, "product.infz")
        product_inf_file = os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", atak_version, "product.inf")
        os.makedirs(os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", atak_version), exist_ok=True)

    if os.path.exists(product_infz_file):
        os.remove(product_infz_file)

    query = db.session.query(Packages)
    if atak_version:
        query = query.where(Packages.atak_version == atak_version)
    packages = db.session.execute(query).all()

    with zipfile.ZipFile(product_infz_file, mode='a', compression=zipfile.ZIP_DEFLATED) as zipf:

        with open(product_inf_file, "w") as inf:
            csv_writer = csv.writer(inf)

            for package in packages:
                package = package[0]
                csv_writer.writerow([package.platform, package.plugin_type, package.package_name,
                                     package.name, package.version, package.revision_code, package.file_name,
                                     package.icon_filename, package.description, package.apk_hash, package.os_requirement,
                                     package.tak_prereq, package.file_size])

                if package.icon:
                    zipf.writestr(package.icon_filename, package.icon)

        zipf.write(product_inf_file, arcname="product.inf")
        os.remove(product_inf_file)


@packages_blueprint.route('/api/packages/repositories.inf')
def get_repository_inf():
    cert = verify_client_cert()
    if not cert:
        return '', 401

    versions = Packages.query.distinct(Packages.atak_version).where(Packages.atak_version is not None).all()
    if not versions:
        return "", 404

    response = ""
    for version in versions:
        if version.atak_version:
            response += version.atak_version + "\n"

    return response, 200


@packages_blueprint.route('/api/packages', methods=['POST'])
@auth_required()
@roles_accepted("administrator")
def add_package():
    form = PackageForm()
    if not form.validate():
        return jsonify({'success': False, 'errors': form.errors}), 400

    os.makedirs(os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages"), exist_ok=True)

    apk_filename = secure_filename(request.files['apk'].filename)

    request.files['apk'].save(os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", apk_filename))

    package = Packages()
    package.from_wtform(form)

    existing_package = db.session.execute(db.session.query(Packages).filter_by(package_name=package.package_name, atak_version=package.atak_version)).scalar()
    if existing_package:
        logger.warning(f"{package.name} version {package.version} for ATAK {package.atak_version} is already on the server and will be removed")

        if os.path.exists(os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", existing_package.file_name)):
            os.remove(os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", existing_package.file_name))

        db.session.delete(existing_package)

    try:
        db.session.add(package)
        db.session.commit()
    except sqlalchemy.exc.IntegrityError:
        db.session.rollback()
        db.session.execute(sqlalchemy.update(Packages).where(Packages.package_name == package.package_name).values(**package.serialize()))
        db.session.commit()

    create_product_infz(form.atak_version.data)

    return jsonify({'success': True})


@packages_blueprint.route('/api/packages', methods=['PATCH'])
@auth_required()
@roles_accepted("administrator")
def edit_package():
    form = PackageUpdateForm(formdata=ImmutableMultiDict(request.json))
    if not form.validate():
        return jsonify({'success': False, 'errors': form.errors}), 400

    package = db.session.execute(db.session.query(Packages).filter_by(package_name=form.package_name.data)).first()
    if not package:
        return jsonify({'success': False, 'error': f"{form.package_name.data} not found"}), 404

    package = package[0]
    package.install_on_enrollment = form.install_on_enrollment.data
    package.install_on_connection = form.install_on_connection.data
    package.publish_time = datetime.datetime.now(datetime.timezone.utc)

    db.session.execute(sqlalchemy.update(Packages).where(Packages.package_name == form.package_name.data).values(**package.serialize()))
    db.session.commit()

    return jsonify({'success': True})


@packages_blueprint.route('/api/packages', methods=['DELETE'])
@auth_required()
@roles_accepted("administrator")
def delete_package():
    package_name = request.args.get("package_name")
    if not package_name:
        return jsonify({'success': False, 'error': 'Please provide the package name of the plugin to delete'}), 400

    atak_version = None

    query = db.session.query(Packages)
    query = search(query, Packages, 'package_name')
    if request.args.get('atak_version'):
        atak_version = bleach.clean(request.args.get("atak_version"))
        query = query.where(Packages.atak_version == atak_version)

    package = db.session.execute(query).first()
    if not package:
        return jsonify({'success': False, 'error': f'Unknown package name: {package_name}'}), 404

    package = package[0]
    if os.path.exists(os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", package.file_name)):
        os.remove(os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", package.file_name))

    db.session.delete(package)
    db.session.commit()

    create_product_infz(atak_version)

    return jsonify({'success': True})



import hashlib
import json
import time
import traceback

import jwt
import os
from dataclasses import dataclass

from opentakserver.extensions import db, logger
from sqlalchemy import String, ForeignKey, Integer, Boolean, BigInteger
from sqlalchemy.orm import Mapped, mapped_column, relationship

from flask import current_app as app


@dataclass
class Token(db.Model):
    __tablename__ = "tokens"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    username: Mapped[str] = mapped_column(String(255), ForeignKey("user.username"), nullable=True, unique=True)
    max_uses: Mapped[int] = mapped_column(Integer, default=None, nullable=True)
    total_uses: Mapped[int] = mapped_column(Integer, default=0, nullable=True)
    creation: Mapped[int] = mapped_column(BigInteger, nullable=False, default=int(time.time()))
    not_before: Mapped[int] = mapped_column(BigInteger, nullable=True)
    expiration: Mapped[int] = mapped_column(BigInteger, nullable=True)
    disabled: Mapped[bool] = mapped_column(Boolean, default=False, nullable=True)
    token_hash: Mapped[str] = mapped_column(String(255), nullable=False, unique=True)

    user = relationship("User", back_populates="tokens")

    def to_json(self, hash=False) -> dict:
        json_token = {
            "sub": self.username,
            "iat": self.creation,
            "iss": "OpenTAKServer",
            "aud": "OpenTAKServer"
        }

        if self.max_uses or not hash:
            json_token["max"] = self.max_uses

        if self.not_before or not hash:
            json_token["nbf"] = self.not_before

        if self.expiration or not hash:
            json_token["exp"] = self.expiration

        return json_token

    def hash_token(self, token: str = None) -> str:
        sha256 = hashlib.sha256()

        if token:
            sha256.update(token.encode())
            token_hash = sha256.hexdigest()
        else:
            sha256.update(json.dumps(self.to_json(True)).encode())
            token_hash = sha256.hexdigest()
            self.token_hash = token_hash

        return token_hash

    def generate_token(self) -> str or None:
        if not self.username:
            return None

        token = {
            "sub": self.username,
            "iat": self.creation,
            "iss": "OpenTAKServer",
            "aud": "OpenTAKServer"
        }

        if self.max_uses:
            token["max"] = self.max_uses

        if self.not_before:
            token["nbf"] = self.not_before

        if self.expiration:
            token["exp"] = self.expiration

        with open(os.path.join(app.config.get("OTS_CA_FOLDER"), "certs", "opentakserver", "opentakserver.nopass.key"), "rb") as key:
            encoded_token = jwt.encode(token, key.read(), algorithm="RS256")
            return encoded_token

    @staticmethod
    def verify_token(token: str) -> bool:
        with open(os.path.join(app.config.get("OTS_CA_FOLDER"), "certs", "opentakserver", "opentakserver.pub"), "r") as key:
            try:
                # Will raise InvalidTokenError on bad signature, expired, or before the nbf date
                decoded_token: dict = jwt.decode(token, key.read(), algorithms=["RS256"], audience="OpenTAKServer")

                sha256 = hashlib.sha256()
                sha256.update(json.dumps(decoded_token).encode())
                token_hash = sha256.hexdigest()

                token_from_db = db.session.query(Token).filter_by(token_hash=token_hash).first()
                if not token_from_db:
                    logger.error(f"Token not in db: {token_hash}")
                    return False

                if token_from_db.disabled:
                    logger.error("Token disabled")
                    return False

                if "max" in decoded_token.keys() and token_from_db.total_uses >= decoded_token["max"]:
                    logger.error(f"Too many uses for token {token_hash}")
                    return False

                token_from_db.total_uses += 1
                db.session.add(token_from_db)
                db.session.commit()

                return True

            except jwt.exceptions.InvalidTokenError as e:
                logger.error(f"Invalid token: {e}")
                logger.debug(traceback.format_exc())
                return False
            except BaseException as e:
                logger.error(f"Failed to decode token: {e}")
                logger.debug(traceback.format_exc())
                return False



from flask_babel import gettext

from flask import Blueprint, jsonify, session, current_app as app

from opentakserver.extensions import babel

language_api = Blueprint('language_api', __name__)


@language_api.route("/api/language/test")
def test_lang():
    return jsonify({"success": True, "message": gettext("This is a test")})


@language_api.route('/api/language')
def get_languages():
    return jsonify(app.config.get("OTS_LANGUAGES"))


@language_api.route('/api/language/<lang_code>', methods=["PUT"])
def set_language(lang_code):
    if not lang_code:
        session['language'] = 'en'

    else:
        lang_codes = []
        for translation in babel.list_translations():
            lang_codes.append(translation.language)

        if lang_code in lang_codes:
            session['language'] = lang_code
            return jsonify({"success": True})
        else:
            return jsonify({"success": False, "error": gettext(f"{lang_code} is not a supported language")}), 400



import base64
import json
import os
import random
import re
import socket
import traceback
import uuid
from xml.etree.ElementTree import Element, SubElement, tostring, fromstring, ParseError
import datetime
from threading import Thread

import bleach
import sqlalchemy
from flask import Flask
from flask_ldap3_login import AuthenticationResponseStatus
from flask_security import verify_password
from flask_socketio import SocketIO

from bs4 import BeautifulSoup
import pika
from meshtastic import mesh_pb2, portnums_pb2, BROADCAST_NUM, mqtt_pb2
from pika.channel import Channel
from sqlalchemy import select, update, insert

from opentakserver.extensions import db, ldap_manager
from opentakserver.functions import datetime_from_iso8601_string, iso8601_string_from_datetime
from opentakserver.models.Chatrooms import Chatroom
from opentakserver.models.EUD import EUD
from opentakserver.models.Group import Group
from opentakserver.models.GroupUser import GroupUser
from opentakserver.models.Meshtastic import MeshtasticChannel
from opentakserver.models.Mission import Mission
from opentakserver.models.CasEvac import CasEvac
from opentakserver.models.GeoChat import GeoChat
from opentakserver.models.ChatroomsUids import ChatroomsUids
from opentakserver.models.MissionChange import MissionChange, generate_mission_change_cot
from opentakserver.models.MissionUID import MissionUID
from opentakserver.models.Team import Team


class ClientController(Thread):
    def __init__(self, address: str, port: int, sock: socket, logger, app: Flask, is_ssl: bool, socketio: SocketIO):
        Thread.__init__(self)
        self.address = address
        self.port = port
        self.sock = sock
        self.logger = logger
        self.shutdown = False
        self.sock.settimeout(1.0)
        self.app = app
        self.db = db
        self.is_ssl = is_ssl
        self.socketio = socketio
        self.bound_queues = []

        self.user = None

        # Device attributes
        self.uid = None
        self.device = None
        self.os = None
        self.platform = None
        self.version = None
        self.callsign = None
        self.phone_number = None
        self.battery = None
        self.groups = {}
        self.device_inserted = False
        self.is_authenticated = False

        # Location Attributes
        self.latitude = 0
        self.longitude = 0
        self.ce = 0
        self.hae = 0
        self.le = 0
        self.course = 0
        self.speed = 0
        self.location_source = None
        self.common_name = None
        self.user = None

        # In case the RabbitMQ channel or connection drops, cached_messages will hold message until the channel is open again
        self.cached_messages = []

        if self.is_ssl:
            try:
                self.sock.do_handshake()
                for c in self.sock.getpeercert()['subject']:
                    if c[0][0] == 'commonName':
                        self.common_name = c[0][1]
                        self.logger.debug("Got common name {}".format(self.common_name))

                        with self.app.app_context():
                            self.user = self.app.security.datastore.find_user(username=self.common_name)
            except BaseException as e:
                logger.warning("Failed to do handshake: {}".format(e))
                self.logger.error(traceback.format_exc())
                self.close_connection()

        # RabbitMQ
        try:
            rabbit_credentials = pika.PlainCredentials(self.app.config.get("OTS_RABBITMQ_USERNAME"), self.app.config.get("OTS_RABBITMQ_PASSWORD"))
            rabbit_host = self.app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
            self.rabbit_connection = pika.SelectConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials), self.on_connection_open)
            self.rabbit_channel: Channel | None = None
            # Start the pika ioloop in a thread or else it blocks and we can't receive any CoT messages
            self.iothread = Thread(target=self.rabbit_connection.ioloop.start, name="IOLOOP")
            self.iothread.daemon = True
            self.iothread.start()
            self.is_consuming = False
        except BaseException as e:
            self.logger.error("Failed to connect to rabbitmq: {}".format(e))
            return

    def on_connection_open(self, connection):
        self.rabbit_connection.channel(on_open_callback=self.on_channel_open)
        self.rabbit_connection.add_on_close_callback(self.on_close)

    def on_channel_open(self, channel: Channel):
        self.logger.debug(f"Opening RabbitMQ channel for {self.callsign or self.address}")
        self.rabbit_channel = channel
        # Remove the on_channel_close callback in case this channel is being reopened
        self.rabbit_channel.callbacks.clear()
        self.rabbit_channel.add_on_close_callback(self.on_channel_close)

        for message in self.cached_messages:
            self.route_cot(message)

        self.cached_messages.clear()

    def on_channel_close(self, channel: Channel, error):
        self.logger.error(f"RabbitMQ channel closed for {self.callsign}, shut it down")
        if self.rabbit_connection and not self.rabbit_connection.is_closing and not self.rabbit_connection.is_closed:
            self.rabbit_connection.close()

        self.shutdown = True
        self.sock.shutdown(socket.SHUT_RDWR)
        self.sock.close()

    def on_close(self, connection, error):
        try:
            if self.rabbit_connection:
                self.rabbit_connection.ioloop.stop()
                self.rabbit_connection = None
        except BaseException as e:
            self.logger.error(f"Failed to close ioloop: {e}")
        self.logger.info("Connection closed for {}: {}".format(self.address, error))

    def on_message(self, unused_channel, basic_deliver, properties, body):
        try:
            body = json.loads(body)
            if body['uid'] != self.uid:
                self.sock.send(body['cot'].encode())
        except BaseException as e:
            self.logger.error(f"{self.callsign}: {e}, closing socket")
            self.close_connection()
            self.logger.error(traceback.format_exc())

    def handle_auth(self, auth: str):
        self.logger.debug(auth)
        if auth:
            auth = BeautifulSoup(auth, 'xml')
        if self.is_ssl and not self.is_authenticated and (auth or self.common_name):
            user = None
            with self.app.app_context():
                if auth:
                    cot = auth.find('cot')
                    if cot:
                        username = cot.attrs['username']
                        password = cot.attrs['password']
                        uid = cot.attrs['uid']

                        if self.app.config.get("OTS_ENABLE_LDAP"):
                            result = ldap_manager.authenticate(username, password)

                            if result.status == AuthenticationResponseStatus.success:
                                # Keep this import here to avoid a circular import when OTS is started
                                from opentakserver.blueprints.ots_api.ldap_api import save_user

                                self.user = save_user(result.user_dn, result.user_id, result.user_info,
                                                      result.user_groups)

                                try:
                                    eud = self.db.session.execute(self.db.session.query(EUD).filter_by(uid=uid)).first()[0]
                                    self.logger.debug("Associating EUD uid {} to user {}".format(eud.uid, self.user.username))
                                    eud.user_id = self.user.id
                                    self.db.session.commit()
                                except:
                                    self.logger.debug("This is a new eud: {} {}".format(uid, self.user.username))
                                    eud = EUD()
                                    eud.uid = uid
                                    eud.user_id = self.user.id
                                    eud.callsign = self.callsign
                                    self.db.session.add(eud)
                                    self.db.session.commit()

                            else:
                                self.close_connection()
                                return

                        else:
                            user = self.app.security.datastore.find_user(username=username)
                elif self.common_name:
                    user = self.app.security.datastore.find_user(username=self.common_name)

                if not user:
                    self.logger.warning("User {} does not exist".format(self.common_name))
                    self.close_connection()
                    return
                elif not user.active:
                    self.logger.warning("User {} is deactivated, disconnecting".format(username))
                    self.close_connection()
                    return
                elif self.common_name:
                    self.logger.info("{} is ID'ed by cert".format(user.username))
                    self.is_authenticated = True
                    self.user = user
                elif verify_password(password, user.password):
                    self.logger.info("Successful login from {}".format(username))
                    self.is_authenticated = True
                    self.user = user
                    try:
                        eud = self.db.session.execute(self.db.session.query(EUD).filter_by(uid=uid)).first()[0]
                        self.logger.debug("Associating EUD uid {} to user {}".format(eud.uid, user.username))
                        eud.user_id = user.id
                        self.db.session.commit()
                    except:
                        self.logger.debug("This is a new eud: {} {}".format(uid, user.username))
                        eud = EUD()
                        eud.uid = uid
                        eud.user_id = user.id
                        self.db.session.add(eud)
                        self.db.session.commit()

                else:
                    self.logger.warning("Wrong password for user {}".format(username))
                    self.close_connection()
                    return

    def handle_cot(self, cot):
        self.logger.debug(cot)
        event = BeautifulSoup(cot, 'xml').find("event")

        # If this client is connected via ssl, make sure they're authenticated
        # before accepting any data from them
        if self.is_ssl and not self.is_authenticated:
            self.logger.warning("EUD isn't authenticated, ignoring")
            return

        if self.pong(event):
            return

        if event and not self.uid:
            self.parse_device_info(event)

        self.route_cot(event)

    # Yes this is super janky
    def run(self):
        cot = ""

        while not self.shutdown:
            try:
                if self.common_name and not self.is_authenticated:
                    self.handle_auth("")

                data = self.sock.recv(65536)
                if not data:
                    # Occurs when an EUD disconnects
                    self.logger.warning("No Data Closing connection to {}".format(self.address))
                    self.shutdown = True
                    self.close_connection()
                    break

                cot += data.decode("utf-8")
                cot_list = re.split("</event>|</auth>", cot)

                if len(cot_list) < 2:
                    continue

                for c in cot_list:
                    try:
                        if "<event" in c:
                            fromstring(c + "</event>")
                            self.handle_cot(c + "</event>")
                        elif "<auth>" in c:
                            fromstring(c + "</auth>")
                            self.handle_auth(c + "</auth>")
                    except ParseError as e:
                        self.logger.error(f"Failed to parse: {e}")
                        cot = c
                        break

                cot = ""

            except TimeoutError:
                if self.shutdown:
                    self.logger.warning("Timeout Closing connection to {}".format(self.address))
                    self.close_connection()
                    break
                else:
                    continue
            except OSError as e:
                self.logger.warning("OSError, stopping")
                self.logger.error(str(e))
                self.logger.error(traceback.format_exc())
                # Client disconnected abruptly, either ATAK crashed, lost network connectivity, the battery died, etc
                return
            except (ConnectionError, ConnectionResetError) as e:
                self.logger.info(f"Closing connection {e}")
                self.close_connection()
                break

    def close_connection(self):
        self.unbind_rabbitmq_queues()
        self.send_disconnect_cot()

        if self.rabbit_channel and not self.rabbit_channel.is_closing and not self.rabbit_channel.is_closed:
            try:
                self.rabbit_channel.close()
            except ValueError as e:
                self.rabbit_connection.ioloop.stop()
                self.rabbit_connection.close()

        if not self.shutdown:
            self.shutdown = True
            self.sock.shutdown(socket.SHUT_RDWR)
            self.sock.close()

    def stop(self):
        self.close_connection()
        self.shutdown = True

    def pong(self, event):
        if event.attrs.get('type') == 't-x-c-t':
            now = datetime.datetime.now(datetime.timezone.utc)
            stale = now + datetime.timedelta(seconds=10)

            cot = Element('event', {'how': 'h-g-i-g-o', 'type': 't-x-c-t-r', 'version': '2.0',
                                      'uid': "{}-pong".format(self.uid), 'start': now, 'time': now, 'stale': stale})
            SubElement(cot, 'point', {'ce': '9999999', 'le': '9999999', 'hae': '0', 'lat': '0',
                                                'lon': '0'})

            try:
                self.sock.send(event.encode())
                return True
            except BaseException as e:
                self.logger.error(e)
                self.logger.debug(traceback.format_exc())

        return False

    def parse_device_info(self, event):
        link = event.find('link')
        fileshare = event.find('fileshare')

        # EUDs running the Meshtastic and dmrcot plugins can relay messages from their RF networks to the server
        # so we want to use the UID of the "off grid" EUD, not the relay EUD
        takv = event.find('takv')
        if takv:
            uid = event.attrs.get('uid')
        else:
            return

        contact = event.find('contact')

        # Only assume it's an EUD if it's got a <takv> tag
        if takv and contact and uid and not uid.endswith('ping') and (self.user or not self.is_ssl):
            self.uid = uid
            device = takv.attrs['device'] if 'device' in takv.attrs else ""
            operating_system = takv.attrs['os'] if 'os' in takv.attrs else ""
            platform = takv.attrs['platform'] if 'platform' in takv.attrs else ""
            version = takv.attrs['version'] if 'version' in takv.attrs else ""

            if 'callsign' in contact.attrs:
                self.callsign = contact.attrs['callsign']

                # Declare a RabbitMQ Queue for this uid and join the 'dms' and 'cot' exchanges
                if self.rabbit_channel and self.rabbit_channel.is_open and platform != "OpenTAK ICU" and platform != "Meshtastic" and platform != "DMRCOT":

                    self.logger.debug(f"Declaring queue for {self.callsign} {self.uid}")
                    self.rabbit_channel.queue_declare(queue=self.callsign)
                    self.rabbit_channel.queue_declare(queue=self.uid)

                    with self.app.app_context():
                        group_memberships = db.session.execute(db.session.query(GroupUser).filter_by(user_id=self.user.id, direction=Group.OUT)).all()
                        if not group_memberships or not self.is_ssl:
                            self.logger.debug(f"{self.callsign} doesn't belong to any groups, adding them to the __ANON__ group")
                            self.rabbit_channel.queue_bind(exchange="groups", queue=self.uid, routing_key="__ANON__.OUT")
                            if {"exchange": "groups", "routing_key": "__ANON__.OUT", "queue": self.uid} not in self.bound_queues:
                                self.bound_queues.append({"exchange": "groups", "routing_key": "__ANON__.OUT", "queue": self.uid})

                        elif group_memberships and self.is_ssl:
                            for membership in group_memberships:
                                membership = membership[0]
                                self.rabbit_channel.queue_bind(exchange="groups", queue=self.uid, routing_key=f"{membership.group.name}.OUT")

                                if {"exchange": "groups", "routing_key": f"{membership.group.name}.OUT", "queue": self.uid} not in self.bound_queues:
                                    self.bound_queues.append({"exchange": "groups", "routing_key": f"{membership.group.name}.OUT", "queue": self.uid})

                        self.rabbit_channel.queue_bind(exchange='missions', routing_key="missions", queue=self.uid)
                        if {"exchange": "missions", "routing_key": "missions", "queue": self.uid} not in self.bound_queues:
                            self.bound_queues.append({"exchange": "groups", "routing_key": "__ANON__.OUT", "queue": self.uid})

                        # The DMs queue also binds by callsign since the <dest> tag in CoT messages can be by callsign instead of UID
                        self.rabbit_channel.queue_bind(exchange='dms', queue=self.uid, routing_key=self.uid)
                        self.rabbit_channel.queue_bind(exchange='dms', queue=self.callsign, routing_key=self.callsign)

                        if {"exchange": "dms", "routing_key": self.uid, "queue": self.uid} not in self.bound_queues:
                            self.bound_queues.append({"exchange": "dms", "routing_key": self.uid, "queue": self.uid})

                        if {"exchange": "dms", "routing_key": self.callsign, "queue": self.callsign} not in self.bound_queues:
                            self.bound_queues.append({"exchange": "dms", "routing_key": self.callsign, "queue": self.callsign})

                        self.rabbit_channel.basic_consume(queue=self.callsign, on_message_callback=self.on_message, auto_ack=True)
                        self.rabbit_channel.basic_consume(queue=self.uid, on_message_callback=self.on_message, auto_ack=True)

            if 'phone' in contact.attrs and contact.attrs['phone']:
                self.phone_number = contact.attrs['phone']

            with self.app.app_context():
                __group = event.find('__group')
                team = Team()

                if __group:
                    team.name = bleach.clean(__group.attrs['name'])

                    try:
                        chatroom = self.db.session.execute(select(Chatroom).filter(Chatroom.name == team.name)).first()[0]
                        team.chatroom_id = chatroom.id
                    except TypeError:
                        chatroom = None

                    try:
                        self.db.session.add(team)
                        self.db.session.commit()
                    except sqlalchemy.exc.IntegrityError:
                        self.db.session.rollback()
                        team = self.db.session.execute(select(Team).filter(Team.name == __group.attrs['name'])).first()[0]
                        if not team.chatroom_id and chatroom:
                            team.chatroom_id = chatroom.id
                            self.db.session.execute(update(Team).filter(Team.name == chatroom.id).values(chatroom_id=chatroom.id))

                try:
                    eud = self.db.session.execute(select(EUD).filter_by(uid=uid)).first()[0]
                except:
                    eud = EUD()

                eud.uid = uid
                if self.callsign:
                    eud.callsign = self.callsign
                if device:
                    eud.device = device

                eud.os = operating_system
                eud.platform = platform
                eud.version = version
                eud.phone_number = self.phone_number
                eud.last_event_time = datetime_from_iso8601_string(event.attrs['start'])
                eud.last_status = 'Connected'
                eud.user_id = self.user.id if self.user else None

                # Set a Meshtastic ID for TAK EUDs to be identified by in the Meshtastic network
                if not eud.meshtastic_id and eud.platform != "Meshtastic":
                    meshtastic_id = '{:x}'.format(int.from_bytes(os.urandom(4), 'big'))
                    while len(meshtastic_id) < 8:
                        meshtastic_id = "0" + meshtastic_id
                    eud.meshtastic_id = int(meshtastic_id, 16)
                elif not eud.meshtastic_id and eud.platform == "Meshtastic":
                    try:
                        eud.meshtastic_id = int(takv.attrs['meshtastic_id'], 16)
                    except:
                        meshtastic_id = '{:x}'.format(int.from_bytes(os.urandom(4), 'big'))
                        while len(meshtastic_id) < 8:
                            meshtastic_id = "0" + meshtastic_id
                        eud.meshtastic_id = int(meshtastic_id, 16)

                # Get the Meshtastic device's mac address or generate a random one for TAK EUDs
                if takv and 'macaddr' in takv.attrs:
                    eud.meshtastic_macaddr = takv.attrs['macaddr']
                else:
                    eud.meshtastic_macaddr = base64.b64encode(os.urandom(6)).decode('ascii')

                if __group:
                    eud.team_id = team.id
                    eud.team_role = bleach.clean(__group.attrs['role'])

                try:
                    self.db.session.add(eud)
                    self.db.session.commit()
                except sqlalchemy.exc.IntegrityError:
                    self.db.session.rollback()
                    self.db.session.execute(update(EUD).where(EUD.uid == eud.uid).values(**eud.serialize()))
                    self.db.session.commit()

                self.send_meshtastic_node_info(eud)
                self.socketio.emit('eud', eud.to_json(), namespace='/socket.io')

    def send_meshtastic_node_info(self, eud):
        if self.app.config.get("OTS_ENABLE_MESHTASTIC") and eud.platform != "Meshtastic":
            user_info = mesh_pb2.User()
            setattr(user_info, "id", "!{:x}".format(eud.meshtastic_id))
            user_info.long_name = eud.callsign
            # Use the last 4 characters of the UID as the short name
            user_info.short_name = eud.uid[-4:]
            user_info.hw_model = mesh_pb2.HardwareModel.PRIVATE_HW

            # Note to future self: The Meshtastic firmware expects a User payload when the Portnum is NodeInfo
            # DO NOT SEND A NODEINFO PAYLOAD!
            encoded_message = mesh_pb2.Data()
            encoded_message.portnum = portnums_pb2.NODEINFO_APP
            encoded_message.payload = user_info.SerializeToString()

            message_id = random.getrandbits(32)
            mesh_packet = mesh_pb2.MeshPacket()
            mesh_packet.id = message_id

            try:
                setattr(mesh_packet, "from", int(eud.meshtastic_id, 16))
            except BaseException as e:
                setattr(mesh_packet, "from", eud.meshtastic_id)
            mesh_packet.to = BROADCAST_NUM
            mesh_packet.want_ack = False
            mesh_packet.hop_limit = 3
            mesh_packet.decoded.CopyFrom(encoded_message)

            service_envelope = mqtt_pb2.ServiceEnvelope()
            service_envelope.packet.CopyFrom(mesh_packet)
            service_envelope.gateway_id = "OpenTAKServer"

            channels = self.db.session.execute(select(MeshtasticChannel).filter(MeshtasticChannel.uplink_enabled is True))
            for channel in channels:
                channel = channel[0]
                service_envelope.channel_id = channel
                routing_key = f"{self.app.config.get('OTS_MESHTASTIC_TOPIC')}.2.e.{channel.name}.outgoing"
                self.rabbit_channel.basic_publish(exchange='amq.topic', routing_key=routing_key, body=service_envelope.SerializeToString(),
                             properties=pika.BasicProperties(expiration=self.app.config.get("OTS_RABBITMQ_TTL")))
                self.logger.debug("Published message to " + routing_key)

    def unbind_rabbitmq_queues(self):
        if self.uid and self.rabbit_channel and not self.rabbit_channel.is_closing and not self.rabbit_channel.is_closed:
            self.rabbit_channel.queue_unbind(queue=self.uid, exchange="missions", routing_key="missions")
            self.rabbit_channel.queue_unbind(queue=self.uid, exchange="groups")
            with self.app.app_context():
                missions = db.session.execute(db.session.query(Mission)).all()
                for mission in missions:
                    self.rabbit_channel.queue_unbind(queue=self.uid, exchange="missions", routing_key=f"missions.{mission[0].name}")
                    self.logger.debug(f"Unbound {self.uid} from mission.{mission[0].name}")

            for bind in self.bound_queues:
                self.rabbit_channel.queue_unbind(exchange=bind['exchange'], queue=bind['queue'], routing_key=bind['routing_key'])

    def send_disconnect_cot(self):
        if self.uid:
            now = datetime.datetime.now(datetime.timezone.utc)
            stale = (datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(seconds=10))

            event = Element('event', {'how': 'h-g-i-g-o', 'type': 't-x-d-d', 'version': '2.0',
                                      'uid': str(uuid.uuid4()), 'start': iso8601_string_from_datetime(now),
                                      'time': iso8601_string_from_datetime(now), 'stale': iso8601_string_from_datetime(stale)})
            point = SubElement(event, 'point', {'ce': '9999999', 'le': '9999999', 'hae': '0', 'lat': '0',
                                                'lon': '0'})
            detail = SubElement(event, 'detail')
            link = SubElement(detail, 'link', {'relation': 'p-p', 'uid': self.uid, 'type': 'a-f-G-U-C'})
            flow_tags = SubElement(detail, '_flow-tags_', {'TAK-Server-f1a8159ef7804f7a8a32d8efc4b773d0': iso8601_string_from_datetime(now)})

            message = json.dumps({'uid': self.uid, 'cot': tostring(event).decode('utf-8')})
            if self.rabbit_channel and not self.rabbit_channel.is_closing and not self.rabbit_channel.is_closed and self.user:
                with self.app.app_context():
                    group_query = self.db.session.query(GroupUser).filter_by(user_id=self.user.id, direction=Group.OUT, enabled=True)
                    groups = self.db.session.execute(group_query).all()
                    for group in groups:
                        group = group[0]
                        self.logger.debug(f"Publishing to group {group.group.name}.{group.direction}")
                        self.rabbit_channel.basic_publish(exchange="groups", routing_key=f"{group.group.name}.{group.direction}", body=message)
            elif self.rabbit_channel and not self.rabbit_channel.is_closing and not self.rabbit_channel.is_closed:
                self.rabbit_channel.basic_publish(exchange="groups", routing_key=f"__ANON__.{Group.OUT}", body=message)

            with self.app.app_context():
                self.db.session.execute(update(EUD).filter(EUD.uid == self.uid).values(last_status='Disconnected', last_event_time=now))
                self.db.session.commit()

        self.logger.info('{} disconnected'.format(self.address))

    def route_cot(self, event):
        if not event:
            return

        if not self.rabbit_channel or not self.rabbit_channel.is_open:
            self.cached_messages.append(event)
            self.logger.error("RabbitMQ channel is closed, not publishing cot")
            return

        # Route all CoTs to the cot exchange for cot_parser and any plugins to receive
        self.rabbit_channel.basic_publish(exchange='cot', body=json.dumps({'uid': self.uid, 'cot': str(event)}), routing_key='',
                                          properties=pika.BasicProperties(expiration=self.app.config.get("OTS_RABBITMQ_TTL")))

        mission_changes = []
        destinations = event.find_all('dest')
        if destinations:

            for destination in destinations:
                creator = event.find("creator")
                creator_uid = self.uid
                if creator and "uid" in creator.attrs:
                    creator_uid = creator.attrs['uid']

                # ATAK and WinTAK use callsign, iTAK uses uid
                if 'callsign' in destination.attrs and destination.attrs['callsign']:
                    self.rabbit_channel.basic_publish(exchange='dms', routing_key=destination.attrs['callsign'],
                                                      body=json.dumps({'uid': self.uid, 'cot': str(event)}))

                # iTAK uses its own UID in the <dest> tag when sending CoTs to a mission so we don't send those to the dms exchange
                elif 'uid' in destination.attrs and destination['uid'] != self.uid:
                    self.rabbit_channel.basic_publish(exchange='dms', routing_key=destination.attrs['uid'],
                                                      body=json.dumps({'uid': self.uid, 'cot': str(event)}))

                # For data sync missions
                elif 'mission' in destination.attrs:
                    with self.app.app_context():
                        mission = self.db.session.execute(
                            self.db.session.query(Mission).filter_by(name=destination.attrs['mission'])).first()

                        if not mission:
                            self.logger.error(f"No such mission found: {destination.attrs['mission']}")
                            return

                        mission = mission[0]
                        self.rabbit_channel.basic_publish("missions",
                                                          routing_key=f"missions.{destination.attrs['mission']}",
                                                          body=json.dumps({'uid': self.uid, 'cot': str(event)}))

                        mission_uid = self.db.session.execute(
                            self.db.session.query(MissionUID).filter_by(uid=event.attrs['uid'])).first()

                        if not mission_uid:
                            mission_uid = MissionUID()
                            mission_uid.uid = event.attrs['uid']
                            mission_uid.mission_name = destination.attrs['mission']
                            mission_uid.timestamp = datetime_from_iso8601_string(event.attrs['start'])
                            mission_uid.creator_uid = creator_uid
                            mission_uid.cot_type = event.attrs['type']

                            color = event.find('color')
                            icon = event.find('usericon')
                            point = event.find('point')
                            contact = event.find('contact')

                            if color and 'argb' in color.attrs:
                                mission_uid.color = color.attrs['argb']
                            elif color and 'value' in color.attrs:
                                mission_uid.color = color.attrs['value']
                            if icon:
                                mission_uid.iconset_path = icon['iconsetpath']
                            if point:
                                mission_uid.latitude = float(point.attrs['lat'])
                                mission_uid.longitude = float(point.attrs['lon'])
                            if contact:
                                mission_uid.callsign = contact.attrs['callsign']

                            try:
                                self.db.session.add(mission_uid)
                                self.db.session.commit()
                            except sqlalchemy.exc.IntegrityError:
                                self.db.session.rollback()
                                self.db.session.execute(update(MissionUID).values(**mission_uid.serialize()))

                            mission_change = MissionChange()
                            mission_change.isFederatedChange = False
                            mission_change.change_type = MissionChange.ADD_CONTENT
                            mission_change.mission_name = destination.attrs['mission']
                            mission_change.timestamp = datetime_from_iso8601_string(event.attrs['start'])
                            mission_change.creator_uid = creator_uid
                            mission_change.server_time = datetime_from_iso8601_string(event.attrs['start'])
                            mission_change.mission_uid = event.attrs['uid']

                            change_pk = self.db.session.execute(insert(MissionChange).values(**mission_change.serialize()))
                            self.db.session.commit()

                            body = {'uid': self.app.config.get("OTS_NODE_ID"), 'cot': tostring(generate_mission_change_cot(destination.attrs['mission'], mission, mission_change, cot_event=event)).decode('utf-8')}
                            mission_changes.append({"mission": mission.name, "message": body})
                            self.rabbit_channel.basic_publish("missions", routing_key=f"missions.{mission.name}", body=json.dumps(body))

        if not destinations and not self.is_ssl:
            # Publish all CoT messages received by TCP and that have no destination to the __ANON__ group
            self.rabbit_channel.basic_publish(exchange='groups', routing_key="__ANON__.OUT", body=json.dumps({'uid': self.uid, 'cot': str(event)}),
                                              properties=pika.BasicProperties(expiration=self.app.config.get("OTS_RABBITMQ_TTL")))
            return

        if not destinations:
            with self.app.app_context():
                group_memberships = db.session.execute(db.session.query(GroupUser).filter_by(user_id=self.user.id, direction=Group.IN, enabled=True)).all()
                if not group_memberships:
                    # Default to the __ANON__ group if the user doesn't belong to any IN groups
                    self.rabbit_channel.basic_publish(exchange='groups', routing_key="__ANON__.OUT", body=json.dumps({'uid': self.uid, 'cot': str(event)}),
                                                      properties=pika.BasicProperties(expiration=self.app.config.get("OTS_RABBITMQ_TTL")))

                for membership in group_memberships:
                    membership = membership[0]
                    self.rabbit_channel.basic_publish(exchange='groups', routing_key=f"{membership.group.name}.{Group.OUT}", body=json.dumps({'uid': self.uid, 'cot': str(event)}),
                                                      properties=pika.BasicProperties(expiration=self.app.config.get("OTS_RABBITMQ_TTL")))

        if mission_changes:
            for change in mission_changes:
                self.rabbit_channel.basic_publish("missions", routing_key=f"missions.{change['mission']}",
                                                  body=json.dumps(change['message']))



from __future__ import annotations

from flask import current_app as app
from abc import abstractmethod

from flask import Flask, Blueprint, url_for

from opentakserver.extensions import logger
from opentakserver.plugins.BasePlugin import BasePlugin


class Plugin(BasePlugin):
    """
    Generic plugin
    """

    def __init__(self):
        self._app: Flask | None = None
        self._config = {}
        self.metadata = {}
        self.name = ""
        self.distro = ""
        self.routes = []

    group = "opentakserver.plugin"
    blueprint: Blueprint | None = None

    @abstractmethod
    def activate(self, app: Flask, enabled: bool) -> None: ...

    @abstractmethod
    def stop(self) -> None: ...

    @abstractmethod
    def get_info(self) -> dict | None: ...

    @abstractmethod
    def load_metadata(self) -> {}: ...

    def has_no_empty_params(self, rule):
        defaults = rule.defaults if rule.defaults is not None else ()
        arguments = rule.arguments if rule.arguments is not None else ()
        return len(defaults) >= len(arguments)

    def get_plugin_routes(self, url_prefix: str):
        for rule in app.url_map.iter_rules():
            # Filter out rules we can't navigate to in a browser
            # and rules that require parameters
            if ("GET" in rule.methods or "POST" in rule.methods) and self.has_no_empty_params(rule):
                url = url_for(rule.endpoint, **(rule.defaults or {}))
                if url.startswith(url_prefix) and url not in self.routes:
                    self.routes.append(url)



import datetime
import hashlib
import json
import os
import time
import traceback
import uuid
from urllib.parse import urlparse

import flask
import jwt

import bleach
import sqlalchemy.exc
from bs4 import BeautifulSoup
from flask import Blueprint, request, current_app as app, jsonify, Response
from flask_babel import gettext
from flask_security import current_user, hash_password, verify_password
from sqlalchemy import update, insert, or_
from werkzeug.utils import secure_filename
import pika

from xml.etree.ElementTree import tostring, Element, fromstring, SubElement

from opentakserver.blueprints.marti_api.marti_api import verify_client_cert
from opentakserver.functions import iso8601_string_from_datetime, datetime_from_iso8601_string
from opentakserver.extensions import db, logger
from opentakserver.models.CoT import CoT
from opentakserver.models.EUD import EUD
from opentakserver.models.Group import Group
from opentakserver.models.GroupMission import GroupMission
from opentakserver.models.GroupUser import GroupUser
from opentakserver.models.Mission import Mission
from opentakserver.models.MissionChange import MissionChange, generate_mission_change_cot
from opentakserver.models.MissionContent import MissionContent
from opentakserver.models.MissionContentMission import MissionContentMission
from opentakserver.models.MissionInvitation import MissionInvitation
from opentakserver.models.MissionLogEntry import MissionLogEntry
from opentakserver.models.MissionRole import MissionRole
from opentakserver.models.MissionUID import MissionUID
from opentakserver.models.Team import Team
from opentakserver.models.user import User

mission_marti_api = Blueprint('mission_marti_api', __name__)


# Only allow access to the mission/data sync API over SSL/port 8443 with a valid client cert.
# nginx will proxy the cert in a header called X-Ssl-Cert by default. This is configurable in ots_https and with the
# OTS_SSL_CERT_HEADER option in config.yml
#@mission_marti_api.before_request
#def verify_client_cert_before_request():
#    if not verify_client_cert():
#        return jsonify({'success': False, 'error': 'Missing or invalid client certificate'}), 400


def verify_token() -> dict | bool:
    token = request.headers.get('Authorization')
    if not token or "Bearer" not in token:
        return False

    token = token.replace("Bearer ", "")

    with open(os.path.join(app.config.get("OTS_CA_FOLDER"), "certs", "opentakserver", "opentakserver.pub"), "r") as key:
        try:
            return jwt.decode(token, key.read(), algorithms=["RS256"])
        except BaseException as e:
            logger.error("Failed to validate mission token: {}".format(e))
            logger.debug(traceback.format_exc())
            return False


# iTAK sucks and doesn't send a token for some reason...
def verify_itak_certificate(mission_name: str = None, mission_guid: str = None) -> MissionRole | flask.Response:
    # Get the username from the client cert forwarded by nginx
    cert = verify_client_cert()
    username = cert.get_subject().commonName

    # Check that the user exists
    user = db.session.execute(db.session.query(User).filter_by(username=username)).first()
    if not user:
        return jsonify({'success': False, 'error': f'User {username} not found'}), 401
    user = user[0]

    # Check that the user owns this EUD
    eud_uid = request.args.get('creatorUid')
    if not eud_uid:
        return jsonify({'success': False, 'error': 'Invalid creatorUid'}), 400

    eud = db.session.execute(db.session.query(EUD).filter_by(uid=eud_uid, user_id=user.id)).first()
    if not eud:
        return jsonify({'success': False, 'error': f'User {username} does not own EUD {eud_uid}'}), 401
    eud = eud[0]

    if not mission_name and mission_guid:
        mission = db.session.execute(db.session.query(Mission).filter_by(guid=mission_guid)).first()
        if not mission:
            return jsonify({'success': False, 'error': f'Invalid mission GUID: {mission_guid}'}), 404
        mission_name = mission[0].name

    # Check that the EUD is subscribed to this mission
    mission_role = db.session.execute(db.session.query(MissionRole).filter_by(clientUid=eud.uid, username=username,
                                                                              mission_name=mission_name)).first()
    if not mission_role:
        logger.error(f"Access denied {username} {mission_name} {eud_uid}")
        return jsonify({'success': False, 'error': 'Access Denied'}), 403

    return mission_role


def check_permission(mission_name: str = None, mission_guid: str = None):
    if "iTAK" not in request.user_agent.string:
        token = verify_token()
        if mission_name and (not token or token['MISSION_NAME'] != mission_name):
            return jsonify({'success': False, 'error': 'Missing or invalid token'}), 401
        elif mission_guid and (not token or token['MISSION_GUID'] != mission_guid):
            return jsonify({'success': False, 'error': 'Missing or invalid token'}), 401
    else:
        cert_is_valid = verify_itak_certificate(mission_name, mission_guid)
        if isinstance(cert_is_valid, flask.Response):
            return cert_is_valid

    return True


def generate_token(mission: Mission, eud_uid: str):
    """
    jti: Unique UUID for the token
    iat: Time token was issued. Can be used to invalidate a token if it was issued before a security event occurred
    sub: The thing the token identifies, the EUD's UID in this case. Used to verify EUD roles, ie MISSION_SUBSCRIBER, MISSION_OWNER, or MISSION_READ_ONLY
    MISSION_NAME: The mission this token is for
    MISSION_GUID: The guid of the mission this token is for

    :param mission:
    :param eud_uid:
    :return: string token
    """
    payload = {'jti': str(uuid.uuid4()), 'iat': int(time.time()), 'sub': eud_uid,
               'iss': urlparse(request.base_url).hostname, 'MISSION_NAME': mission.name,
               'MISSION_GUID': mission.guid}

    server_key = open(os.path.join(app.config.get("OTS_CA_FOLDER"), "certs", "opentakserver",
                                   "opentakserver.nopass.key"), "r")

    token = jwt.encode(payload, server_key.read(), algorithm="RS256")
    server_key.close()

    return token


def generate_new_mission_cot(mission: Mission) -> Element:
    event = Element("event", {"type": "t-x-m-n", "how": "h-g-i-g-o", "version": "2.0", "uid": str(uuid.uuid4()),
                              "start": iso8601_string_from_datetime(datetime.datetime.now(datetime.timezone.utc)),
                              "time": iso8601_string_from_datetime(datetime.datetime.now(datetime.timezone.utc)),
                              "stale": iso8601_string_from_datetime(
                                  datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(hours=1))})
    SubElement(event, "point", {'ce': '9999999', 'le': '9999999', 'hae': '0', 'lat': '0', 'lon': '0'})
    detail = SubElement(event, "detail")
    SubElement(detail, "mission", {'type': Mission.CREATE, 'tool': mission.tool, 'name': mission.name,
                                   'guid': mission.guid, 'authorUid': mission.creator_uid})

    return event


def generate_invitation_cot(mission: Mission, uid: str, cot_type: str = "t-x-m-i", delete: bool = False) -> Element:
    """
    Generates an invitation (t-x-m-i) or role change (t-x-m-r) cot
    :param mission:
    :param uid:
    :param cot_type:
    :param delete:
    :return:
    """

    event = Element("event", {"type": cot_type, "how": "h-g-i-g-o", "version": "2.0", "uid": str(uuid.uuid4()),
                              "start": iso8601_string_from_datetime(datetime.datetime.now(datetime.timezone.utc)),
                              "time": iso8601_string_from_datetime(datetime.datetime.now(datetime.timezone.utc)),
                              "stale": iso8601_string_from_datetime(
                                  datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(hours=1))})

    SubElement(event, "point", {"ce": "9999999", "le": "9999999", "hae": "0", "lat": "0", "lon": "0"})
    detail = SubElement(event, "detail")
    mission_tag = SubElement(detail, "mission", {"type": Mission.INVITE, "tool": mission.tool, "name": mission.name,
                                                 "guid": mission.guid, "authorUid": mission.creator_uid,
                                                 "token": generate_token(mission, uid)})

    if not delete:
        role = SubElement(mission_tag, "role", {"type": mission.default_role})
        permissions = SubElement(role, "permissions")

        if mission.default_role == MissionRole.MISSION_SUBSCRIBER:
            SubElement(permissions, "permission", {"type": MissionRole.MISSION_READ})
            SubElement(permissions, "permission", {"type": MissionRole.MISSION_WRITE})
        elif mission.default_role == MissionRole.MISSION_OWNER:
            SubElement(permissions, "permission", {"type": MissionRole.MISSION_MANAGE_FEEDS})
            SubElement(permissions, "permission", {"type": MissionRole.MISSION_SET_PASSWORD})
            SubElement(permissions, "permission", {"type": MissionRole.MISSION_WRITE})
            SubElement(permissions, "permission", {"type": MissionRole.MISSION_MANAGE_LAYERS})
            SubElement(permissions, "permission", {"type": MissionRole.MISSION_UPDATE_GROUPS})
            SubElement(permissions, "permission", {"type": MissionRole.MISSION_DELETE})
            SubElement(permissions, "permission", {"type": MissionRole.MISSION_SET_ROLE})
            SubElement(permissions, "permission", {"type": MissionRole.MISSION_READ})
        else:
            SubElement(permissions, "permission", {"type": MissionRole.MISSION_READ})

    return event


def generate_mission_delete_cot(mission: Mission) -> Element:
    event = Element("event", {"type": "t-x-m-d", "how": "h-g-i-g-o", "version": "2.0", "uid": str(uuid.uuid4()),
                              "start": iso8601_string_from_datetime(datetime.datetime.now(datetime.timezone.utc)),
                              "time": iso8601_string_from_datetime(datetime.datetime.now(datetime.timezone.utc)),
                              "stale": iso8601_string_from_datetime(datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(hours=1))})
    SubElement(event, "point", {'ce': '9999999', 'le': '9999999', 'hae': '0', 'lat': '0', 'lon': '0'})
    detail = SubElement(event, "detail")
    SubElement(detail, "mission", {'type': Mission.DELETE, 'tool': mission.tool, 'name': mission.name,
                                             'guid': mission.guid, 'authorUid': mission.creator_uid})

    return event


@mission_marti_api.route('/Marti/api/missions/guid/<mission_guid>')
def get_mission_by_guid(mission_guid: str):
    permission_granted = check_permission(mission_guid=mission_guid)
    if isinstance(permission_granted, flask.Response):
        return permission_granted

    password = request.args.get('password')
    mission = db.session.execute(db.session.query(Mission).filter_by(guid=mission_guid)).first()
    if not mission:
        return jsonify({'success': False, 'error': gettext(u'No mission found with guid: %(mission_guid)s', mission_guid=mission_guid)}), 404
    mission = mission[0]

    if mission.password_protected and not verify_password(password, mission.password):
        return jsonify({'success': False, 'error': gettext(u'Invalid password')}), 401

    return jsonify({'version': "3", 'type': 'Mission', 'data': [mission.to_json()], 'nodeId': app.config.get("OTS_NODE_ID")})


@mission_marti_api.route('/Marti/api/missions')
def get_missions():
    cert = verify_client_cert()
    if not cert:
        return '', 401

    username = cert.get_subject().commonName
    user = app.security.datastore.find_user(username=username)

    password_protected = request.args.get('passwordProtected', False)

    tool = request.args.get('tool')
    if tool:
        tool = bleach.clean(tool)

    default_role = request.args.get('defaultRole')
    if default_role:
        default_role = bleach.clean(default_role).lower() == 'true'

    response = {
        'version': 3, 'type': 'Mission', 'data': [], 'nodeId': app.config.get('OTS_NODE_ID')
    }

    try:
        query = db.session.query(Mission)

        # Let admins see all missions
        if not user.has_role('administrator'):
            group_filters = []
            groups = db.session.execute(db.session.query(GroupUser).filter_by(user_id=user.id, direction=Group.IN)).scalars()
            for group in groups:
                group_filters.append(GroupMission.group_id == group.group.id)
            if group_filters:
                query = query.outerjoin(GroupMission).where(or_(*group_filters))
            else:
                # If a user isn't in a group, only show them __ANON__ missions
                query.outerjoin(GroupMission).where(GroupMission.group_id == 1)

        missions = db.session.execute(query).scalars()
        for mission in missions:
            if not password_protected and mission.password_protected:
                continue
            if tool and tool.lower() != "public" and mission.tool != tool:
                continue
            response['data'].append(mission.to_marti_json())

    except BaseException as e:
        logger.error(f"Failed to get missions: {e}")
        logger.debug(traceback.format_exc())
        return jsonify({'success': False, 'error': str(e)}), 500

    return jsonify(response)


@mission_marti_api.route('/Marti/api/missions/all/invitations', methods=['GET'])
@mission_marti_api.route('/Marti/api/missions/invitations', methods=['GET'])
def all_invitations():
    if 'clientUid' in request.args and request.args.get('clientUid'):
        client_uid = bleach.clean(request.args.get('clientUid'))
    else:
        return '', 200

    response = {
        'version': "3", 'type': 'Mission', 'data': [],
        'nodeId': app.config.get("OTS_NODE_ID"), 'messages': []
    }

    invitations = db.session.execute(db.session.query(MissionInvitation).filter_by(client_uid=client_uid)).all()

    for invitation in invitations:
        response['data'].append(invitation[0].mission_name)

    return jsonify(response)


@mission_marti_api.route('/Marti/api/missions/<mission_name>', methods=['PUT', 'POST'])
def put_mission(mission_name: str):
    """ Used by the Data Sync plugin to create or change a mission """
    cert = verify_client_cert()
    if not cert:
        return jsonify({"success": False, "error": gettext(u"Missions are only supported on SSL connections")}), 400

    username = cert.get_subject().commonName
    user = app.security.datastore.find_user(username=username)

    new_mission = True

    if not mission_name or not request.args.get('creatorUid'):
        return jsonify({'success': False, 'error': gettext(u'Please provide a mission name and creatorUid')}), 400

    eud = db.session.execute(db.session.query(EUD).filter_by(uid=request.args.get('creatorUid'))).first()
    if not eud:
        return jsonify({'success': False, 'error': gettext(u"Invalid creatorUid: %(creator_uid)s",
                                                           creator_uid=request.args.get('creatorUid'))}), 400
    eud = eud[0]

    password = None
    mission = db.session.execute(db.session.query(Mission).filter_by(name=mission_name)).first()
    if mission:
        mission = mission[0]
        new_mission = False
    else:
        mission = Mission()
        mission.name = bleach.clean(mission_name)
        if 'password' in request.args and request.args.get('password'):
            password = hash_password(request.args.get('password'))

    mission.creator_uid = mission.creator_uid or request.args.get('creatorUid') or mission.creator_uid or None
    mission.description = request.args.get('description') or mission.description or None
    mission.tool = request.args.get('tool') or mission.tool or "public"
    mission.group = request.args.get('group') or mission.group or "__ANON__"
    mission.default_role = request.args.get('defaultRole') or mission.default_role or MissionRole.MISSION_SUBSCRIBER
    mission.password = password or mission.password or None
    mission.password_protected = (mission.password is not None)
    mission.guid = mission.guid or str(uuid.uuid4())
    mission.create_time = mission.create_time or datetime.datetime.now(datetime.timezone.utc)

    try:
        db.session.add(mission)
        # Will raise IntegrityError if the mission exists, meaning we should update it
        db.session.commit()

        groups = request.args.get('group')
        if groups:
            for group_name in groups.split(","):
                group = db.session.execute(db.session.query(Group).filter_by(name=group_name)).scalar()
                if not group:
                    return jsonify({"success": False, "error": gettext("Group not found: %(group_name)s", group_name=group_name)}), 404
                group_mission = GroupMission()
                group_mission.group_id = group.id
                group_mission.mission_name = mission.name
                db.session.add(group_mission)
        else:
            # Default to the __ANON__ group
            group_mission = GroupMission()
            group_mission.group_id = 1
            group_mission.mission_name = mission.name
            db.session.add(group_mission)

        db.session.commit()

        if new_mission:
            mission_role = MissionRole()
            mission_role.clientUid = mission.creator_uid
            mission_role.username = eud.user.username if eud.user else "anonymous"
            mission_role.createTime = datetime.datetime.now(datetime.timezone.utc)
            mission_role.role_type = MissionRole.MISSION_OWNER
            mission_role.mission_name = mission_name
            db.session.add(mission_role)

            mission_change = MissionChange()
            mission_change.isFederatedChange = False
            mission_change.change_type = MissionChange.CREATE_MISSION
            mission_change.mission_name = mission_name
            mission_change.timestamp = mission.create_time
            mission_change.creator_uid = mission.creator_uid
            mission_change.server_time = mission.create_time

            db.session.add(mission_change)
            db.session.commit()

            event = generate_new_mission_cot(mission)

            rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
            rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
            rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
            channel = rabbit_connection.channel()

            groups = db.session.execute(db.session.query(GroupUser).filter_by(user_id=user.id, enabled=True)).scalars()
            for group in groups:
                channel.basic_publish(exchange="groups", routing_key=f"{group.group.name}.{group.direction}", body=json.dumps({'uid': app.config.get("OTS_NODE_ID"), 'cot': tostring(event).decode('utf-8')}))

            channel.close()
            rabbit_connection.close()
    except sqlalchemy.exc.IntegrityError:
        # Mission exists, needs updating
        db.session.rollback()
        db.session.execute(update(Mission).where(Mission.name == mission_name).values(**mission.serialize()))
        db.session.commit()
        return jsonify({'version': "3", 'type': 'Mission', 'data': [mission.to_json()], 'nodeId': app.config.get("OTS_NODE_ID")})
    except BaseException as e:
        logger.error(f"Failed to add mission: {e}")
        logger.debug(traceback.format_exc())
        return jsonify({'success': False, 'error': gettext("Failed to add mission: %(e)s", e=str(e))}), 500

    token = generate_token(mission, mission.creator_uid)
    mission_json = mission.to_json()
    mission_json['token'] = token

    if new_mission:
        mission_json['ownerRole'] = MissionRole.OWNER_ROLE

    response = {'version': "3", 'type': 'Mission', 'data': [mission_json], 'nodeId': app.config.get("OTS_NODE_ID")}

    if new_mission:
        return jsonify(response), 201
    else:
        return jsonify(response), 200


@mission_marti_api.route('/Marti/api/missions/<mission_name>', methods=['GET'])
def get_mission(mission_name: str):
    """ Used by the Data Sync plugin to get a feed's metadata """
    if not mission_name:
        return jsonify({'success': False, 'error': gettext(u'Invalid mission name')}), 400

    mission_name = bleach.clean(mission_name)

    try:
        mission = db.session.execute(db.session.query(Mission).filter_by(name=mission_name)).first()
        if not mission:
            return jsonify({'success': False, 'error': gettext('Mission %(mission_name)S not found', mission_name=mission_name)}), 404

        return jsonify({'version': "3", 'type': 'Mission', 'data': [mission[0].to_json()], 'nodeId': app.config.get("OTS_NODE_ID")})
    except BaseException as e:
        logger.error(f"Failed to get mission: {e}")
        logger.debug(traceback.format_exc())
        return jsonify({'success': False, 'error': str(e)}), 400


@mission_marti_api.route('/Marti/api/missions/<mission_name>', methods=['DELETE'])
def delete_mission(mission_name: str):
    """ Used by the Data Sync plugin to delete a feed """

    # ATAK sends a creatorUid param, but we ignore it in favor of the UID in the signed JWT token that ATAK also sends.
    creator_uid = request.args.get('creatorUid')

    if "iTAK" not in request.user_agent.string:
        token = verify_token()
        if not token or token['MISSION_NAME'] != mission_name:
            return jsonify({'success': False, 'error': gettext(u'Missing or invalid token')}), 401
        eud_uid = token['sub']
    else:
        # cert_is_valid will either be True or flask.Response. If it's flask.Response it indicates an error
        role = verify_itak_certificate(mission_name)
        if isinstance(role, flask.Response):
            return role
        eud_uid = role.clientUid

    mission = db.session.execute(db.session.query(Mission).filter_by(name=mission_name)).first()
    if mission:
        mission = mission[0]

        can_delete = False

        # Check if the UID in the token has the MISSION_OWNER role for this mission. If not, it can't delete the mission
        for role in mission.roles:
            if role.clientUid == eud_uid and role.role_type == MissionRole.MISSION_OWNER:
                can_delete = True
                break

        if not can_delete:
            return jsonify({'success': False, 'error': gettext(u'Only mission owners can delete missions')}), 403

        db.session.delete(mission)
        db.session.commit()

        rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
        rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
        rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
        channel = rabbit_connection.channel()
        channel.basic_publish(exchange="missions", routing_key="missions",
                              body=json.dumps({'uid': app.config.get("OTS_NODE_ID"),
                                               'cot': tostring(generate_mission_delete_cot(mission)).decode('utf-8')}))
        channel.close()
        rabbit_connection.close()

        return jsonify({'success': True})
    else:
        return jsonify({'success': False, 'error': gettext(u'Mission %(mission_name)s not found', mission_name=mission_name)}), 404


@mission_marti_api.route('/Marti/api/missions/<mission_name>/password', methods=['PUT', 'DELETE'])
def set_password(mission_name: str):
    """ Used by the Data Sync plugin to add a password to a feed """
    if "iTAK" not in request.user_agent.string:
        token = verify_token()
        if not token or token['MISSION_NAME'] != mission_name:
            return jsonify({'success': False, 'error': gettext(u'Missing or invalid token')}), 401
        eud_uid = token['sub']
    else:
        # cert_is_valid will either be True or flask.Response. If it's flask.Response it indicates an error
        role = verify_itak_certificate(mission_name)
        if isinstance(role, flask.Response):
            return role
        eud_uid = role.clientUid

    if request.method == 'PUT' and 'password' not in request.args:
        return jsonify({'success': False, 'error': gettext(u'Please provide the password')}), 400

    role = db.session.execute(db.session.query(MissionRole).filter_by(mission_name=mission_name, clientUid=eud_uid)).first()
    if not role or role[0].role_type != MissionRole.MISSION_OWNER:
        return jsonify({'success': False, 'error': gettext(u"You do not have permission to change this mission's password")}), 403

    creator_uid = request.args.get('creatorUid')

    if request.method == 'PUT':
        password = hash_password(request.args.get('password'))
        db.session.execute(update(Mission).where(Mission.name == mission_name).values(password=password, password_protected=True))
    elif request.method == 'DELETE':
        db.session.execute(update(Mission).where(Mission.name == mission_name).values(password=None, password_protected=False))
    db.session.commit()

    return jsonify({'success': True})


@mission_marti_api.route('/Marti/api/missions/<mission_name>/invite/<invitation_type>/<invitee>', methods=['PUT'])
def invite(mission_name: str, invitation_type: str, invitee: str):
    permission_granted = check_permission(mission_name)
    if isinstance(permission_granted, flask.Response):
        return permission_granted

    mission = db.session.execute(db.session.query(Mission).filter_by(name=mission_name)).first()
    if not mission:
        return jsonify({'success': False, 'error': gettext(u"Mission %(mission_name)s not found", mission_name=mission_name)}), 404

    mission = mission[0]

    invitation = MissionInvitation()
    invitation.mission_name = mission_name

    if invitation_type.lower() == "clientuid":
        eud = db.session.execute(db.session.query(EUD).filter_by(uid=invitee)).first()
        if not eud:
            return jsonify({'success': False, 'error': gettext(u"No EUD found with UID %(invitee)s", invitee=invitee)}), 404
        invitation.client_uid = invitee

    elif invitation_type == "callsign":
        eud = db.session.execute(db.session.query(EUD).filter_by(callsign=invitee)).first()
        if not eud:
            return jsonify({'success': False, 'error': gettext(u"No EUD found with callsign %(invitee)s", invitee=invitee)}), 404
        invitation.callsign = invitee

    elif invitation_type.lower() == "username":
        eud = db.session.execute(db.session.query(User).filter_by(username=invitee)).first()
        if not eud:
            return jsonify({'success': False, 'error': gettext(u"No user found with username %(invitee)s", invitee=invitee)}), 404
        invitation.username = invitee

    elif invitation_type == "group":
        group = db.session.execute(db.session.query(Group).filter_by(group_name=invitee)).first()
        if not group:
            return jsonify({'success': False, 'error': gettext(u"No group found: %(invitee)s", invitee=invitee)}), 404
        invitation.group_name = invitee

    elif invitation_type == "team":
        team = db.session.execute(db.session.query(Team).filter_by(name=invitee)).first()
        if not team:
            return jsonify({'success': False, 'error': gettext(u"Team not found: %(invitee)s", invitee=invitee)}), 404
        invitation.team_name = invitee

    db.session.add(invitation)
    db.session.commit()

    event = generate_invitation_cot(mission, invitee)
    rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
    rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
    rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
    channel = rabbit_connection.channel()
    channel.basic_publish(exchange="dms", routing_key=invitee, body=json.dumps({"uid": app.config.get("OTS_NODE_ID"),
                                                                                "cot": tostring(event).decode('utf-8')}))
    channel.close()
    rabbit_connection.close()

    return '', 200


@mission_marti_api.route('/Marti/api/missions/<mission_name>/invite/<invitation_type>/<invitee>', methods=['DELETE'])
def delete_invitation(mission_name: str, invitation_type: str, invitee: str):
    permission_granted = check_permission(mission_name)
    if isinstance(permission_granted, flask.Response):
        return permission_granted

    mission = db.session.execute(db.session.query(Mission).filter_by(name=mission_name)).first()
    if not mission:
        return jsonify({'success': False, 'error': gettext(u"Mission %(mission_name)s not found", mission_name=mission_name)}), 404

    mission = mission[0]

    if invitation_type.lower() not in ['clientuid', 'callsign', 'username', 'group', 'team']:
        return jsonify({'success': False, 'error': gettext(u"Invalid invitation type: %(invitation_type)s", invitation_type=invitation_type)}), 400

    # Doing it like this because I can select an EUD, user, group, or team and automatically
    # get all of their invitations
    query = db.session.query(MissionInvitation).where(MissionInvitation.mission_name == mission_name)
    if invitation_type.lower() == 'clientuid':
        query = query.where(MissionInvitation.client_uid == invitee)
    elif invitation_type.lower() == 'callsign':
        query = query.where(MissionInvitation.callsign == invitee)
    elif invitation_type.lower() == 'username':
        query = query.where(MissionInvitation.username == invite)
    elif invitation_type.lower() == 'group':
        query = query.where(MissionInvitation.group == invite)
    elif invitation_type.lower() == 'team':
        query = query.where(MissionInvitation.team == invite)

    invitations = db.session.execute(query).scalars()
    for invitation in invitations:
        db.session.delete(invitation)
    db.session.commit()

    return jsonify({'success': True})


@mission_marti_api.route('/Marti/api/missions/<mission_name>/invite', methods=['POST'])
def invite_json(mission_name: str):
    permission_granted = check_permission(mission_name)
    if isinstance(permission_granted, flask.Response):
        return permission_granted

    creator_uid = request.args.get("creatorUid")
    invitees = request.json

    mission = db.session.execute(db.session.query(Mission).filter_by(name=mission_name)).first()
    if not mission:
        return jsonify({'success': False, 'error': gettext(u"Mission not found: %(mission_name)s", mission_name=mission_name)}), 404
    mission = mission[0]

    rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
    rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
    rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
    channel = rabbit_connection.channel()

    for invitee in invitees:
        invitation = MissionInvitation()
        invitation.mission_name = mission_name
        invitation.role = invitee['role']['type']

        if invitee['type'].lower() == 'clientuid':
            invitation.client_uid = invitee['invitee']
        elif invitee['type'].lower() == 'callsign':
            invitation.callsign = invitee['invitee']
        elif invitee['type'].lower() == 'username':
            invitation.username = invitee['invitee']
        elif invitee['type'].lower() == 'group':
            invitation.group = invitee['invitee']
        elif invitee['type'].lower() == 'team':
            invitation.team = invitee['invitee']
        else:
            return jsonify({'success': False, 'error': gettext(u"Invalid invitation type: %(invitation)s", invitation=invitation['type'])}), 400

        db.session.add(invitation)
        db.session.commit()

        event = generate_invitation_cot(mission, invitee['invitee'])

        logger.debug(f"Sending invitation to mission {mission_name} to {invitee['invitee']}")
        channel.basic_publish(exchange="dms", routing_key=invitee['invitee'],
                              body=json.dumps({'uid': app.config['OTS_NODE_ID'], "cot": tostring(event).decode('utf-8')}))

        channel.close()
        rabbit_connection.close()

    return jsonify({'success': True})


@mission_marti_api.route('/Marti/api/missions/<mission_name>/subscriptions/roles')
def mission_roles(mission_name: str):
    permission_granted = check_permission(mission_name)
    if isinstance(permission_granted, flask.Response):
        return permission_granted

    response = {"version": "3", "type": "MissionSubscription", "data": [], "nodeId": app.config.get("OTS_NODE_ID")}
    roles = db.session.execute(db.session.query(MissionRole).filter_by(mission_name=mission_name))
    for role in roles:
        response['data'].append(role[0].to_json())

    return jsonify(response)


@mission_marti_api.route('/Marti/api/missions/guid/<mission_guid>/subscriptions/roles')
def mission_roles_by_guid(mission_guid: str):
    permission_granted = check_permission(mission_guid=mission_guid)
    if isinstance(permission_granted, flask.Response):
        return permission_granted

    response = {"version": "3", "type": "MissionSubscription", "data": [], "nodeId": app.config.get("OTS_NODE_ID")}
    mission = db.session.execute(db.session.query(Mission).filter_by(guid=mission_guid)).first()
    if mission:
        mission = mission[0]
        roles = db.session.execute(db.session.query(MissionRole).filter_by(mission_name=mission.name))
        for role in roles:
            response['data'].append(role[0].to_json())

    return jsonify(response)


@mission_marti_api.route('/Marti/api/missions/<mission_name>/role', methods=['PUT'])
def change_eud_role(mission_name: str):
    """ Used by Data Sync to change EUD mission roles or kick an EUD off of a mission """
    if "iTAK" not in request.user_agent.string:
        token = verify_token()
        if not token or token['MISSION_NAME'] != mission_name:
            return jsonify({'success': False, 'error': gettext(u'Missing or invalid token')}), 401
        eud_uid = token['sub']
    else:
        # cert_is_valid will either be True or flask.Response. If it's flask.Response it indicates an error
        role = verify_itak_certificate(mission_name)
        if isinstance(role, flask.Response):
            return role
        eud_uid = role.clientUid

    mission = db.session.execute(db.session.query(Mission).filter_by(name=mission_name)).first()
    if not mission:
        return jsonify({'success': False, 'error': gettext(u"No such mission found: %(mission_name)s", mission_name=mission_name)}), 404
    mission = mission[0]

    role = db.session.execute(db.session.query(MissionRole).filter_by(clientUid=eud_uid, role_type=MissionRole.MISSION_OWNER)).first()
    if not role:
        return jsonify({'success': False, 'error': gettext(u'Only mission owners can change EUD roles')}), 403

    client_uid = request.args.get('clientUid')
    if not client_uid:
        return jsonify({'success': False, 'error': gettext(u'Please provide a UID')}), 400

    eud = db.session.execute(db.session.query(EUD).filter_by(uid=client_uid)).first()
    if not eud:
        return jsonify({'success': False, 'error': gettext(u'Invalid UID: %(client_uid)s', client_uid=client_uid)}), 400
    eud = eud[0]

    new_role = request.args.get('role')
    if new_role and new_role not in [MissionRole.MISSION_OWNER, MissionRole.MISSION_SUBSCRIBER, MissionRole.MISSION_READ_ONLY]:
        return jsonify({'success': False, 'error': gettext(u"Invalid role: %(new_role)s", new_role=new_role)})
    elif new_role:
        r = db.session.execute(db.session.query(MissionRole).filter_by(clientUid=client_uid, mission_name=mission_name)).all()
        for role in r:
            db.session.delete(role[0])
        db.session.commit()

        role = MissionRole()
        role.clientUid = client_uid
        try:
            role.username = eud.user.username
        except BaseException as e:
            role.username = "anonymous"
        role.createTime = datetime.datetime.now(datetime.timezone.utc)
        role.role_type = new_role
        role.mission_name = mission_name

        db.session.add(role)
        db.session.commit()

        event = generate_invitation_cot(mission, role.clientUid)
        body = {'uid': app.config.get("OTS_NODE_ID"), 'cot': tostring(event).decode('utf-8')}

        rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
        rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
        rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
        channel = rabbit_connection.channel()
        channel.basic_publish(exchange="dms", routing_key=client_uid, body=json.dumps(body))
        channel.close()
        rabbit_connection.close()

    # No new role provided, kick the EUD off the mission
    else:
        old_role = db.session.execute(db.session.query(MissionRole).filter_by(mission_name=mission_name, clientUid=client_uid)).first()
        if old_role:
            db.session.delete(old_role[0])
            db.session.commit()

        event = generate_invitation_cot(mission, client_uid, 't-x-m-r', delete=True)
        body = {'uid': app.config.get("OTS_NODE_ID"), 'cot': tostring(event).decode('utf-8')}

        rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
        rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
        rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
        channel = rabbit_connection.channel()
        channel.basic_publish(exchange="dms", routing_key=client_uid, body=json.dumps(body))
        channel.close()
        rabbit_connection.close()

    return '', 200


@mission_marti_api.route('/Marti/api/missions/guid/<mission_guid>/role')
def get_role_by_guid(mission_guid: str):
    permission_granted = check_permission(mission_guid=mission_guid)
    if isinstance(permission_granted, flask.Response):
        return permission_granted

    mission = db.session.execute(db.session.query(Mission).filter_by(guid=mission_guid)).first()
    if not mission:
        return jsonify({'success': False, 'error': gettext(u'No mission found with guid: %(mission_guid)s', mission_guid=mission_guid)}), 404
    mission = mission[0]

    response = {"version": "3", "type": "com.bbn.marti.sync.model.MissionRole", "data": mission.roles[0].to_json()['role'], "nodeId": app.config.get("OTS_NODE_ID")}

    return jsonify(response)


@mission_marti_api.route('/Marti/api/missions/<mission_name>/subscriptions')
def get_subscriptions(mission_name: str):
    permission_granted = check_permission(mission_name)
    if isinstance(permission_granted, flask.Response):
        return permission_granted

    subscriptions = db.session.execute(db.session.query(MissionRole).filter_by(mission_name=mission_name)).all()
    response = {
        "version": "3", "type": "MissionSubscription", "data": [], "nodeId": app.config.get("OTS_NODE_ID")
    }

    for subscription in subscriptions:
        response['data'].append(subscription[0].clientUid)

    return jsonify(response)


@mission_marti_api.route('/Marti/api/missions/<mission_name>/keywords', methods=['PUT'])
def put_mission_keywords(mission_name):
    permission_granted = check_permission(mission_name)
    if isinstance(permission_granted, flask.Response):
        return permission_granted

    mission = db.session.execute(db.session.query(Mission).filter_by(name=mission_name)).first()
    if not mission:
        return jsonify({'success': False, 'error': gettext(u"Cannot find mission %(mission_name)s", mission_name=mission_name)}), 404
    mission = mission[0]

    new_keywords = request.json
    current_keywords = mission.keywords or []
    for keyword in new_keywords:
        if keyword not in current_keywords:
            current_keywords.append(keyword)

    db.session.execute(update(Mission).where(Mission.name == mission_name).values(keywords=current_keywords))
    db.session.commit()

    return '', 200


@mission_marti_api.route('/Marti/api/missions/<mission_name>/subscription', methods=['PUT'])
@mission_marti_api.route('/Marti/api/missions/guid/<mission_guid>/subscription', methods=['PUT'])
def mission_subscribe(mission_name: str = None, mission_guid: str = None):
    """ Used by the Data Sync plugin to subscribe to a feed """
    cert = verify_client_cert()
    username = cert.get_subject().commonName
    user = app.security.datastore.find_user(username=username)

    if mission_name:
        mission = db.session.execute(db.session.query(Mission).filter_by(name=mission_name)).first()
    elif mission_guid:
        mission = db.session.execute(db.session.query(Mission).filter_by(guid=mission_guid)).first()
    else:
        mission = None

    if not mission and mission_name:
        return jsonify({'success': False, 'error': gettext(u"Cannot find mission %(mission_name)s", mission_name=mission_name)}), 404
    if not mission and mission_guid:
        return jsonify({'success': False, 'error': gettext(u"Cannot find mission %(mission_guid)s", mission_guid=mission_guid)}), 404

    mission = mission[0]
    group_filters = []
    groups = db.session.execute(db.session.query(GroupUser).filter_by(user_id=user.id)).scalars()
    for group in groups:
        group_filters.append(GroupMission.group_id == group.group_id)
    if not group_filters:
        # Default to the __ANON__ group
        group_filters.append(GroupMission.group_id == 1)

    query = db.session.query(GroupMission).filter_by(mission_name=mission_name)
    query.where(or_(*group_filters))
    mission_groups = db.session.execute(query).scalars()

    if not mission_groups:
        return jsonify({"success": False, "error": gettext(u"{username} and mission %(mission_name)s are not in the same group", mission_name=mission_name)}), 403

    response = {
        "version": "3", "type": "com.bbn.marti.sync.model.MissionSubscription", "data": {}, "nodeId": app.config.get("OTS_NODE_ID")
    }

    # And EUD will send a token if it has previously subscribed to the mission
    if 'Authorization' in request.headers:
        token = verify_token()
        if not token or token['MISSION_NAME'] != mission_name:
            return jsonify({'success': False, 'error': gettext(u'Invalid token')}), 400

        eud = db.session.execute(db.session.query(EUD).filter_by(uid=token['sub'])).first()[0]
        uid = token['sub']
        role = db.session.execute(db.session.query(MissionRole).filter_by(mission_name=mission_name, clientUid=uid)).first()

        # If this request has a token but no role in the DB, this EUD was invited and this is its first time subscribing
        if not role:
            role = MissionRole()
            role.clientUid = token['sub']
            role.username = eud.user.username if eud.user else 'anonymous'
            role.createTime = datetime.datetime.now(datetime.timezone.utc)
            role.role_type = mission.default_role
            role.mission_name = token['MISSION_NAME']

            db.session.add(role)
            db.session.commit()
        else:
            role = role[0]

        response['data'] = {
            "token": request.headers.get('Authorization').replace('Bearer ', ''),
            "clientUid": token['sub'],
            "username": role.username,
            "createTime": role.createTime,
            "role": role.to_json()['role'],
        }

    # If no token is sent, this is a new subscription request
    else:
        if "uid" not in request.args:
            return jsonify({'success': False, 'error': gettext(u'Missing UID')}), 400

        uid = bleach.clean(request.args.get("uid"))
        eud = db.session.execute(db.session.query(EUD).filter_by(uid=uid)).first()
        if not eud:
            return jsonify({'success': False, 'error': gettext(u"Invalid UID: %(uid)s", uid=uid)}), 400
        eud = eud[0]

        if mission.password_protected:
            if not verify_password(request.args.get('password', ''), mission.password):
                return jsonify({'success': False, 'error': gettext(u'Invalid password')}), 401

        role = db.session.execute(db.session.query(MissionRole).filter_by(mission_name=mission_name, clientUid=uid)).first()
        if not role:
            role = MissionRole()
            role.clientUid = uid
            role.username = eud.user.username
            role.createTime = datetime.datetime.now(datetime.timezone.utc)
            role.role_type = mission.default_role
            role.mission_name = mission.name

            db.session.add(role)
            db.session.commit()
        else:
            role = role[0]

        token = generate_token(mission, uid)

        response['data'] = {
            "token": token,
            "clientUid": uid,
            "mission": mission.to_json(),
            "username": role.username,
            "createTime": role.createTime,
            "role": role.to_json()['role'],
        }

    rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
    rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
    rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
    channel = rabbit_connection.channel()
    channel.queue_declare(queue=uid)
    channel.queue_bind(queue=uid, exchange="missions", routing_key=f"missions.{mission_name}")
    channel.close()
    rabbit_connection.close()

    # Delete any invitations to this mission for this EUD
    invitations = db.session.execute(db.session.query(MissionInvitation).filter_by(mission_name=mission_name, client_uid=uid)).all()
    for invitation in invitations:
        db.session.delete(invitation[0])
    db.session.commit()

    return jsonify(response), 201


@mission_marti_api.route('/Marti/api/missions/<mission_name>/subscription', methods=['DELETE'])
def mission_unsubscribe(mission_name: str):
    """ Used by the Data Sync plugin to unsubscribe to a feed """
    if "iTAK" not in request.user_agent.string:
        token = verify_token()
        if not token or token['MISSION_NAME'] != mission_name:
            return jsonify({'success': False, 'error': gettext(u'Missing or invalid token')}), 401
        eud_uid = token['sub']
    else:
        # cert_is_valid will either be True or flask.Response. If it's flask.Response it indicates an error
        role = verify_itak_certificate(mission_name)
        if isinstance(role, flask.Response):
            return role
        eud_uid = role.clientUid

    #if "uid" not in request.args:
    #    return jsonify({'success': False, 'error': 'Missing UID'}), 400

    #uid = bleach.clean(request.args.get("uid"))
    role = db.session.execute(db.session.query(MissionRole).filter_by(clientUid=eud_uid, mission_name=mission_name)).first()

    if role:
        db.session.delete(role[0])
        db.session.commit()

    rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
    rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
    rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
    channel = rabbit_connection.channel()
    channel.queue_unbind(queue=eud_uid, exchange="missions", routing_key=f"missions.{mission_name}")
    channel.close()
    rabbit_connection.close()

    return '', 200


@mission_marti_api.route('/Marti/api/missions/<mission_name>/changes', methods=['GET'])
def mission_changes(mission_name):
    permission_granted = check_permission(mission_name)
    if isinstance(permission_granted, flask.Response):
        return permission_granted

    squashed = request.args.get('squashed')
    if squashed:
        squashed = bleach.clean(squashed)

    response = {
        "version": "3", "type": "MissionChange", "data": [], "nodeId": app.config.get("OTS_NODE_ID")
    }

    changes = db.session.execute(db.session.query(MissionChange).filter_by(mission_name=mission_name)).all()
    for change in changes:
        response['data'].append(change[0].to_json())

    return jsonify(response)


@mission_marti_api.route('/Marti/api/missions/logs/entries', methods=['POST'])
def create_log_entry():
    permission_granted = check_permission(request.json['missionNames'][0])
    if isinstance(permission_granted, flask.Response):
        return permission_granted

    mission = db.session.execute(db.session.query(Mission).filter_by(name=request.json['missionNames'][0])).first()
    if not mission:
        return jsonify({'success': False, 'error': gettext(u"Mission not found: %(mission_name)s", mission_name=request.json['missionNames'][0])}), 404

    log_entry = MissionLogEntry()
    log_entry.content = request.json.get('content')
    log_entry.creator_uid = request.json.get('creatorUid')
    log_entry.entry_uid = str(uuid.uuid4())
    log_entry.mission_name = request.json.get('missionNames')[0]
    log_entry.server_time = datetime.datetime.now(datetime.timezone.utc)
    log_entry.dtg = datetime_from_iso8601_string(request.json.get('dtg'))
    log_entry.created = datetime.datetime.now(datetime.timezone.utc)
    log_entry.keywords = request.json.get('keywords')

    db.session.add(log_entry)
    db.session.commit()

    response = {
        'version': '3', 'type': 'com.bbn.marti.sync.model.LogEntry', 'nodeId': app.config.get('OTS_NODE_ID'), 'data': [log_entry.to_json()]
    }

    change_cot = log_entry.generate_cot()
    body = json.dumps({'uid': log_entry.creator_uid, 'cot': tostring(change_cot).decode('utf-8')})

    rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
    rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
    rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
    channel = rabbit_connection.channel()
    channel.basic_publish("missions", routing_key=f"missions.{log_entry.mission_name}", body=body)
    channel.close()
    rabbit_connection.close()

    return jsonify(response), 201


@mission_marti_api.route('/Marti/api/missions/<mission_name>/log', methods=['GET'])
def mission_log(mission_name):
    permission_granted = check_permission(mission_name)
    if isinstance(permission_granted, flask.Response):
        return permission_granted

    mission = db.session.execute(db.session.query(Mission).filter_by(name=mission_name)).first()
    if not mission:
        return jsonify({'success': False, 'error': gettext(u"Mission %(mission_name)s not found", mission_name=mission_name)}), 404
    mission = mission[0]

    response = {
        "version": "3", "type": "com.bbn.marti.sync.model.LogEntry", "data": [],
        "nodeId": app.config.get("OTS_NODE_ID")
    }

    for log in mission.mission_logs:
        response['data'].append(log.to_json())

    return jsonify(response)


@mission_marti_api.route('/Marti/sync/upload', methods=['POST'])
def upload_content():
    """
    Used by the Data Sync plugin when adding files to a mission
    Also used to upload files and data packages

    :return: flask.Response
    """

    cert = verify_client_cert()
    if not cert:
        return jsonify({"success": False, "error": gettext(u"Missing or invalid certificate")}), 400

    username = cert.get_subject().commonName

    file_name = bleach.clean(request.args.get('name')) if 'name' in request.args else None
    keywords = request.args.getlist('keywords')

    if 'creatorUid' in request.args:
        creator_uid = request.args.get('creatorUid')
    # Older versions of iTAK use CreatorUid instead of creatorUid
    elif 'CreatorUid' in request.args:
        creator_uid = request.args.get('CreatorUid')
    else:
        creator_uid = None

    if not file_name:
        return jsonify({'success': False, 'error': gettext(u'File name cannot be blank')}), 400

    filename, extension = os.path.splitext(secure_filename(file_name))

    # In some cases iTAK doesn't include the file extension. If the user agent includes iTAK and there's no extension we'll assume it's a zip
    if not extension and "iTAK" in request.user_agent.string:
        extension = "zip"

    if extension.replace('.', '').lower() not in app.config.get("ALLOWED_EXTENSIONS"):
        logger.error(f"{extension} is not an allowed file extension")
        return jsonify({'success': False, 'error': gettext(u'%(extension)s is not an allowed file extension', extension=extension)}), 400

    file = request.data
    sha256 = hashlib.sha256()
    sha256.update(file)

    content = db.session.execute(db.session.query(MissionContent).filter_by(hash=sha256.hexdigest())).first()
    if not content:
        content = MissionContent()
        content.mime_type = request.content_type
        content.filename = file_name
        content.submission_time = datetime.datetime.now(datetime.timezone.utc)
        content.submitter = username or "anonymous"
        content.uid = str(uuid.uuid4())
        content.creator_uid = creator_uid
        content.size = request.content_length
        content.expiration = -1
        content.keywords = keywords if keywords else []
        content.hash = sha256.hexdigest()
        content_pk = db.session.execute(insert(MissionContent).values(**content.serialize()))
        content_pk = content_pk.inserted_primary_key[0]
        db.session.commit()

    else:
        content = content[0]
        content_pk = content.id

        # For some reason iTAK changes file names to a timestamp with the format YYYYMMDD-HHMMSS so the file name in the DB
        # needs to be updated
        if file_name != content.filename:
            content.filename = file_name
            db.session.add(content)
            db.session.commit()

    # Save the content even if it exists in the database in case it was deleted from disk
    os.makedirs(os.path.join(app.config.get("OTS_DATA_FOLDER"), 'missions'), exist_ok=True)
    with open(os.path.join(app.config.get("OTS_DATA_FOLDER"), 'missions', file_name), 'wb') as f:
        f.write(file)
        f.flush()

    response = {
        "UID": content.uid, "SubmissionDateTime": iso8601_string_from_datetime(content.submission_time), "MIMEType": content.mime_type,
        "SubmissionUser": content.submitter, "PrimaryKey": content_pk, "Hash": content.hash, "CreatorUid": creator_uid,
        "Name": file_name
    }

    return jsonify(response)


@mission_marti_api.route('/Marti/api/sync/metadata/<content_hash>/keywords', methods=['PUT'])
def add_content_keywords(content_hash: str):
    # Not validating if the EUD is subscribed to the mission since we're not given the mission name or GUID
    # Instead, assume this EUD isn't malicious since we require a valid cert in order to get here

    keywords = request.json
    content = db.session.execute(db.session.query(MissionContent).filter_by(hash=content_hash)).first()
    if not content:
        return jsonify({'success': False, 'error': gettext(u"No content found with hash: %(content_hash)s", content_hash=content_hash)}), 400
    content: MissionContent = content[0]
    current_keywords = content.keywords if content.keywords else []

    for keyword in keywords:
        if keyword not in current_keywords:
            current_keywords.append(keyword)
    db.session.execute(update(MissionContent).where(MissionContent.hash == content_hash).values(keywords=current_keywords))
    db.session.commit()

    return '', 200


@mission_marti_api.route('/Marti/api/missions/<mission_name>/contents', methods=['PUT'])
def mission_contents(mission_name: str):
    """ Associates content/files with a mission """
    permission_granted = check_permission(mission_name)
    if isinstance(permission_granted, flask.Response):
        return permission_granted

    body = request.json

    mission = db.session.execute(db.session.query(Mission).filter_by(name=mission_name)).first()
    if not mission:
        logger.error(f"No such mission: {mission_name}")
        return jsonify({'success': False, 'error': gettext(u"No such mission: %(mission_name)s", mission_name=mission_name)}), 404

    mission = mission[0]

    if 'hashes' in body:
        for content_hash in body['hashes']:
            content = db.session.execute(db.session.query(MissionContent).filter_by(hash=content_hash)).first()
            if not content:
                logger.error(f"No such file with hash {content_hash}")
                return jsonify({'success': False, 'error': gettext(u"No such file with hash %(content_hash)s", content_hash=content_hash)}), 404

            content: MissionContent = content[0]

            mission_content_mission = db.session.execute(db.session.query(MissionContentMission).filter_by(mission_content_id=content.id, mission_name=mission_name)).first()
            if not mission_content_mission:
                mission_content_mission = MissionContentMission()
                mission_content_mission.mission_name = mission_name
                mission_content_mission.mission_content_id = content.id

                db.session.add(mission_content_mission)

            mission_change = db.session.execute(db.session.query(MissionChange).filter_by(content_uid=content.uid, mission_name=mission_name)).first()
            if not mission_change:
                mission_change = MissionChange()
                mission_change.isFederatedChange = False
                mission_change.change_type = MissionChange.ADD_CONTENT
                mission_change.content_uid = content.uid
                mission_change.mission_name = mission_name
                mission_change.timestamp = datetime.datetime.now(datetime.timezone.utc)
                mission_change.creator_uid = content.creator_uid
                mission_change.server_time = datetime.datetime.now(datetime.timezone.utc)

                db.session.add(mission_change)

                event = generate_mission_change_cot(mission_name, mission, mission_change, content=content)

                body = json.dumps({'uid': mission_change.creator_uid, 'cot': tostring(event).decode('utf-8')})
                rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
                rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
                rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
                channel = rabbit_connection.channel()
                channel.basic_publish("missions", routing_key=f"missions.{mission_name}", body=body)
                channel.close()
                rabbit_connection.close()

    if 'uids' in body:
        for uid in body['uids']:
            mission_uid = db.session.execute(db.session.query(MissionUID).filter_by(uid=uid, mission_name=mission_name)).first()
            if mission_uid:
                mission_uid = mission_uid[0]
            else:
                mission_uid = MissionUID()

            mission_uid.uid = uid
            mission_uid.timestamp = datetime.datetime.now(datetime.timezone.utc)
            mission_uid.creator_uid = request.args.get('creatorUid')
            mission_uid.mission_name = mission_name

            # iTAK sucks. It sends a CoT and makes a PUT to this endpoint rather than including a <dest mission="mission_name">
            # tag in the CoT. This endpoint finishes before the CoT can be parsed and inserted into the database. In that case
            # we insert a row in the mission_uids table with the CoT data missing, and the parse_point method in
            # cot_controller will fill it in
            cot = db.session.execute(db.session.query(CoT).filter_by(uid=uid)).first()
            if cot:
                cot = cot[0]

                mission_uid.cot_type = cot.type
                mission_uid.latitude = cot.point.latitude
                mission_uid.longitude = cot.point.longitude

                event = BeautifulSoup(cot.xml, 'xml')
                usericon = event.find('usericon')
                color = event.find('color')
                contact = event.find('contact')

                if usericon and 'iconsetpath' in usericon.attrs:
                    mission_uid.iconset_path = usericon.attrs['iconsetpath']
                elif usericon and 'iconsetPath' in usericon.attrs:
                    mission_uid.iconset_path = usericon.attrs['iconsetPath']

                if color and 'argb' in color.attrs:
                    mission_uid.color = color.attrs['argb']
                if color and 'value' in color.attrs:
                    mission_uid.color = color.attrs['value']

                if contact and 'callsign' in contact.attrs:
                    mission_uid.callsign = contact.attrs['callsign']

            mission_change = MissionChange()
            mission_change.isFederatedChange = False
            mission_change.change_type = MissionChange.ADD_CONTENT
            mission_change.mission_name = mission_name
            mission_change.timestamp = datetime.datetime.now(datetime.timezone.utc)
            mission_change.creator_uid = request.args.get('creatorUid')
            mission_change.server_time = datetime.datetime.now(datetime.timezone.utc)
            mission_change.mission_uid = uid

            db.session.execute(insert(MissionChange).values(**mission_change.serialize()))

            event = generate_mission_change_cot(mission_name, mission, mission_change, mission_uid=mission_uid)

            body = json.dumps({'uid': mission_change.creator_uid, 'cot': tostring(event).decode('utf-8')})
            logger.warning(f"{body}")
            rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"),
                                                       app.config.get("OTS_RABBITMQ_PASSWORD"))
            rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
            rabbit_connection = pika.BlockingConnection(
                pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
            channel = rabbit_connection.channel()
            channel.basic_publish("missions", routing_key=f"missions.{mission_name}", body=body)
            channel.close()
            rabbit_connection.close()

            try:
                db.session.add(mission_uid)
                db.session.commit()
            except sqlalchemy.exc.IntegrityError:
                db.session.rollback()
                db.session.execute(update(MissionUID).where(MissionUID.uid == mission_uid.uid).values(**mission_uid.serialize()))
                db.session.commit()

    db.session.commit()

    return jsonify({"version": "3", "type": "Mission", "data": [mission.to_json()], "nodeId": app.config.get("OTS_NODE_ID")})


@mission_marti_api.route('/Marti/api/missions/<mission_name>/contents', methods=['DELETE'])
def delete_content(mission_name: str):
    if "iTAK" not in request.user_agent.string:
        token = verify_token()
        if not token or token['MISSION_NAME'] != mission_name:
            return jsonify({'success': False, 'error': gettext(u'Missing or invalid token')}), 401
        eud_uid = token['sub']
    else:
        # cert_is_valid will either be True or flask.Response. If it's flask.Response it indicates an error
        role = verify_itak_certificate(mission_name)
        if isinstance(role, flask.Response):
            return role
        eud_uid = role.clientUid

    mission = db.session.execute(db.session.query(Mission).filter_by(name=mission_name)).first()
    if not mission:
        return jsonify({'success': False, 'error': gettext(u'Mission %(mission_name)s not found', mission_name=mission_name)}), 404
    mission = mission[0]

    mission_change = MissionChange()
    mission_change.isFederatedChange = False
    mission_change.change_type = MissionChange.REMOVE_CONTENT
    mission_change.mission_name = mission_name
    mission_change.timestamp = datetime.datetime.now(datetime.timezone.utc)
    mission_change.creator_uid = request.args.get('creatorUid')
    mission_change.server_time = datetime.datetime.now(datetime.timezone.utc)

    mission_uid = None
    cot_event = None
    if 'uid' in request.args:
        mission_uid = db.session.execute(db.session.query(MissionUID).filter_by(uid=request.args.get('uid'))).first()
        if not mission_uid:
            return jsonify({'success': False, 'error': gettext(u"UID %(uid)s not found", uid=request.args.get('uid'))}), 404
        else:
            mission_uid = mission_uid[0]
            mission_uid.mission_name = None
            mission_change.mission_uid = mission_uid.uid
        cot_event = db.session.execute(db.session.query(CoT).filter_by(uid=request.args.get("uid"))).first()
        if cot_event:
            cot_event = cot_event[0]
            cot_event.mission_name = None
            db.session.add(cot_event)
            cot_event = BeautifulSoup(cot_event.xml, 'xml').find('event')

    # Files will be kept in the DB so the mission log is correct and on disk in case it gets added back to a mission
    content = None
    if 'hash' in request.args:
        content = db.session.execute(db.session.query(MissionContent).filter_by(hash=request.args.get('hash'))).first()
        if not content:
            return jsonify({'success': False, 'error': gettext(u"No content found with hash %(hash)s", hash=request.args.get('hash'))}), 404
        content = content[0]

        mission_change.content_uid = content.uid

        try:
            mission_content_mission = db.session.execute(db.session.query(MissionContentMission)
                                                         .filter_by(mission_name=mission_name, mission_content_id=content.id)).first()
            if mission_content_mission:
                db.session.delete(mission_content_mission[0])
                db.session.commit()
        except BaseException as e:
            logger.error(f"Failed to delete content with hash {request.args.get('hash')}: {e}")
            logger.debug(traceback.format_exc())
            return jsonify({'success': False, 'error': gettext(u"Failed to delete content with hash %(hash)s: %(e)s", hash=request.args.get('hash'), e=str(e))}), 500

    event = generate_mission_change_cot(eud_uid, mission, mission_change, content=content, mission_uid=mission_uid, cot_event=cot_event)
    body = {'uid': app.config.get("OTS_NODE_ID"), 'cot': tostring(event).decode('utf-8')}

    rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"),
                                               app.config.get("OTS_RABBITMQ_PASSWORD"))
    rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
    rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
    channel = rabbit_connection.channel()
    channel.basic_publish(exchange="missions", routing_key=f"missions.{mission_name}", body=json.dumps(body))
    channel.close()
    rabbit_connection.close()

    db.session.add(mission_change)
    db.session.commit()

    return jsonify({"version": "3", "type": "Mission", "data": [mission.to_json()], "nodeId": app.config.get("OTS_NODE_ID")})


@mission_marti_api.route('/Marti/api/missions/<mission_name>/contents/missionpackage', methods=['PUT'])
def add_content(mission_name):
    client_uid = request.args.get('clientUid')
    if 'Authorization' not in request.headers or not verify_token():
        return jsonify({'success': False, 'error': gettext(u'Missing or invalid token')}), 401

    return '', 200


@mission_marti_api.route('/Marti/api/missions/<mission_name>/cot')
@mission_marti_api.route('/Marti/api/missions/guid/<mission_guid>/cot')
def get_mission_cots(mission_name: str = None, mission_guid: str = None):
    """
    Used by the Data Sync plugin to get all CoTs associated with a feed. Returns the CoTs encapsulated by an
    <events> tag
    """

    permission_granted = check_permission(mission_name, mission_guid)
    if isinstance(permission_granted, flask.Response):
        return permission_granted

    if mission_name:
        cots = db.session.execute(db.session.query(CoT).filter_by(mission_name=mission_name)).all()
    elif mission_guid:
        mission = db.session.execute(db.session.query(Mission).filter_by(guid=mission_guid)).first()
        if mission:
            mission_name = mission[0].name
            cots = db.session.execute(db.session.query(CoT).filter_by(mission_name=mission_name)).all()
        else:
            cots = []
    else:
        cots = []

    events = Element("events")

    for cot in cots:
        events.append(fromstring(cot[0].xml))

    return Response(response=tostring(events).decode('utf-8'), status=200, mimetype='application/xml')


@mission_marti_api.route('/Marti/api/missions/<mission_name>/layers')
def get_mission_layers():
    return ''



"""Added last_meshtastic_publish to EUDs table

Revision ID: 6665d030f187
Revises: 1041fae84708
Create Date: 2025-01-30 14:07:58.600407

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '6665d030f187'
down_revision = '1041fae84708'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('euds', schema=None) as batch_op:
        batch_op.add_column(sa.Column('last_meshtastic_publish', sa.DateTime(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('euds', schema=None) as batch_op:
        batch_op.drop_column('last_meshtastic_publish')


    # ### end Alembic commands ###



from opentakserver.extensions import db
from sqlalchemy import String, Integer, ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship


class Chatroom(db.Model):
    __tablename__ = 'chatrooms'

    id: Mapped[str] = mapped_column(String(255), primary_key=True)
    name: Mapped[str] = mapped_column(String(255))
    group_owner: Mapped[str] = mapped_column(String(255), nullable=True)
    parent: Mapped[str] = mapped_column(String(255), nullable=True)
    geochats = relationship("GeoChat", cascade="all, delete-orphan", back_populates="chatroom")
    chatroom_uid = relationship("ChatroomsUids", cascade="all, delete-orphan", back_populates="chatroom")
    team = relationship("Team", back_populates="chatroom")

    def serialize(self):
        return {
            'name': self.name,
            'group_owner': self.group_owner,
            'parent': self.parent,
        }

    def to_json(self):
        return {
            'name': self.name,
            'group_owner': self.group_owner,
            'parent': self.parent,
            'geochats': [chat.to_json() for chat in self.geochats] if self.geochats else None
        }
